import { EMPTY_OBJ, Lifecycle, NO_OP, isArray, isBrowser, isFunction, isInvalid, isNullOrUndef, isNumber, isObject, isStringOrNumber, isUndefined, throwError } from 'mesh-envoy-shared';

var asyncToGenerator = function (fn) {
  return function () {
    var gen = fn.apply(this, arguments);
    return new Promise(function (resolve, reject) {
      function step(key, arg) {
        try {
          var info = gen[key](arg);
          var value = info.value;
        } catch (error) {
          reject(error);
          return;
        }

        if (info.done) {
          resolve(value);
        } else {
          return Promise.resolve(value).then(function (value) {
            step("next", value);
          }, function (err) {
            step("throw", err);
          });
        }
      }

      return step("next");
    });
  };
};

var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

var createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();









































var toConsumableArray = function (arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
};

var MODULE_MAP = {};


var isMesh = typeof mesh !== 'undefined';

/**
 * patchQueue 还有 update 方法
 */
var patchQueue = [];
function excutePatchQueue() {
	if (isMesh && patchQueue.length > 0) {
		mesh.update(checkAlignBottom(patchQueue));
		patchQueue = [];
	}
}

/**
 * 解决布局问题
 * @param {} patchQueue 
 */
function checkAlignBottom(patchQueue) {
	// MARK: Because snabbdom insert a align bottom elm will make the iOS client render error, so delete the reference node to avoid it
	return patchQueue.map(function (patch) {
		if (patch.type === 'insert' && patch.node.attrs && patch.node.attrs.alignment === 'bottom') {
			var referenceNode = MODULE_MAP[patch.referenceNode];
			var attrs = referenceNode && referenceNode.data.attrs;
			if (!attrs || attrs.alignment !== 'bottom') {
				patch.referenceNode = null;
			}
		}
		return patch;
	});
}

/**
 * help 方法，解决 handler this 指向的问题
 * 因为所有 tag 上的 handler 响应函数都是在全局作用域上执行的，避免去使用 this
 * @param {} handler 
 * @param {} params 
 */
function linkEvent(handler) {
	for (var _len = arguments.length, params = Array(_len > 1 ? _len - 1 : 0), _key2 = 1; _key2 < _len; _key2++) {
		params[_key2 - 1] = arguments[_key2];
	}

	if (isFunction(handler)) {
		return function () {
			for (var _len2 = arguments.length, args = Array(_len2), _key3 = 0; _key3 < _len2; _key3++) {
				args[_key3] = arguments[_key3];
			}

			return handler.apply(undefined, toConsumableArray(params.concat(args)));
		};
	} else {
		if (process.env.NODE_ENV !== 'production') {
			throwError('a valid Component Event handler in linkEvent');
		}
		throwError();
	}
}

/**
 *  发给客户端之前， 清理客户端不需要的 property 
 * @param {*} dom 
 */
function stripDom(dom) {
	var result = {
		tagName: dom.tagName,
		moduleId: dom.moduleId,
		attrs: dom.attrs,
		text: ''
	};

	if (isArray(dom.childNodes)) {
		result.childNodes = dom.childNodes.map(function (child) {
			if (isNullOrUndef(child.data)) {
				return stripDom(child);
			} else {
				result.text += child.data;
			}
		}).filter(function (child) {
			return !isNullOrUndef(child);
		});
	}

	return result;
}

/**
 * 这里保存的 tag 信息是用来检测 tag 的相关关系，保证正确的使用 tag，当客户端新增或者修改了 tag 之后需要来这里添加或修改，当然也可以去掉这一部分，只通过 mesh-envoy-tag 来控制， 这些 tag 的属性可以在 http://dev.kiwiinc.net/kiwi-team/Mesh/wikis/Components/ButtonGroup 这里找到
 */
var normalTags = {};
normalTags = {
	div: {},
	mainpage: {
		child: {
			required: {
				featurelist: normalTags.featurelist
			}
		}
	},
	settingpage: {},
	videopage: {},
	handlerpage: {},
	cameraexposuresettingpanel: {
		attrs: {
			safeShutterSpeed: {
				required: false,
				canModify: true
			}
		}
	},
	slidebar: {
		attrs: {
			value: {
				required: true,
				canModify: true
			},
			step: {
				required: true,
				canModify: false
			}
		},
		child: {
			required: {
				start: normalTags.start,
				end: normalTags.end
			}
		}
	},
	degreebar: {
		attrs: {
			value: {
				required: true,
				canModify: true
			},
			step: {
				required: true,
				canModify: false
			}
		},
		child: {
			required: {
				start: normalTags.start,
				end: normalTags.end
			},
			mid: normalTags.mid
		}
	},
	start: {
		attrs: {
			value: {
				required: true,
				canModify: false
			},
			text: {
				required: true,
				canModify: false
			}
		}
	},
	mid: {
		attrs: {
			text: {
				required: false,
				canModify: false
			}
		}
	},
	end: {
		attrs: {
			value: {
				required: true,
				canModify: false
			},
			text: {
				required: true,
				canModify: false
			}
		}
	},
	couplelabel: {
		attrs: {
			style: {
				required: false,
				canModify: false
			},
			leftContent: {
				required: true,
				canModify: true
			},
			rightContent: {
				required: true,
				canModify: true
			}
		}
	},
	detaillabel: {
		attrs: {
			title: {
				required: true,
				canModify: false
			},
			text: {
				required: false,
				canModify: true
			}
		}
	},
	plainlabel: {
		attrs: {
			text: {
				required: true,
				canModify: true
			}
		}
	},
	ratelabel: {
		attrs: {
			title: {
				required: true,
				canModify: false
			},
			leftUnit: {
				required: true,
				canModify: false
			},
			rightUnit: {
				required: true,
				canModify: false
			},
			leftContent: {
				required: true,
				canModify: true
			},
			rightContent: {
				required: true,
				canModify: true
			}
		}
	},
	iconlabel: {
		attrs: {
			title: {
				required: true,
				canModify: false
			},
			url: {
				required: false,
				canModify: true
			},
			file: {
				required: false,
				canModify: true
			}
		}
	},
	executebutton: {
		attrs: {
			title: {
				required: true,
				canModify: true
			},
			executing: {
				required: true,
				canModify: true
			}
		}
	},
	capturebutton: {
		attrs: {
			type: {
				required: true,
				canModify: true
			},
			title: {
				required: false,
				canModify: true
			}
		}
	},
	detailswitcher: {
		attrs: {
			title: {
				required: true,
				canModify: true
			},
			value: {
				required: true,
				canModify: true
			}
		}
	},
	map: {
		attrs: {
			waypoints: {
				required: false,
				canModify: true
			},
			selected: {
				required: false,
				canModify: true
			},
			type: {
				required: false,
				canModify: false
			}
		}
	},
	waypointmap: {
		attrs: {
			waypoints: {
				required: false,
				canModify: true
			},
			selected: {
				required: false,
				canModify: true
			},
			type: {
				required: false,
				canModify: false
			}
		}
	},
	missionmap: {
		attrs: {
			waypoints: {
				required: false,
				canModify: true
			}
		}
	}
};

var listItemTags = {
	feature: {
		parent: ['featurelist'],
		attrs: {
			text: {
				required: true,
				canModify: false
			},
			icon: {
				required: true,
				canModify: false
			}
		}
	},
	button: {
		parent: ['buttongroup'],
		attrs: {
			text: {
				required: true,
				canModify: true
			},
			type: {
				required: false,
				canModify: false
			}
		}
	},
	switchitem: {
		parent: ['handlerlist'],
		attrs: {
			title: {
				required: true,
				canModify: false
			},
			off: {
				required: true,
				canModify: false
			},
			on: {
				required: true,
				canModify: false
			},
			enable: {
				required: false,
				canModify: true
			},
			value: {
				required: true,
				canModify: true
			}
		}
	},
	detailitem: {
		parent: ['handlerlist'],
		attrs: {
			title: {
				required: true,
				canModify: false
			},
			text: {
				required: true,
				canModify: true
			},
			hidden: {
				required: false,
				canModify: true
			}
		}
	},
	menuitem: {
		parent: ['handlerlist'],
		attrs: {
			text: {
				required: true,
				canModify: true
			},
			style: {
				required: false,
				canModify: true
			}
		}
	},
	stepperitem: {
		parent: ['handlerlist'],
		attrs: {
			title: {
				required: true,
				canModify: false
			},
			value: {
				required: true,
				canModify: true
			},
			index: {
				required: true,
				canModify: true
			},
			hidden: {
				required: false,
				canModify: true
			}
		}
	},
	item: {
		parent: ['horizontallist', 'verticallist'],
		attrs: {
			title: {
				required: false,
				canModify: false
			},
			text: {
				required: false,
				canModify: true
			},
			icon: {
				required: false,
				canModify: false
			}
		}
	}
};

var listTags = {
	featurelist: {
		child: {
			required: {
				feature: listItemTags.feature
			}
		}
	},
	buttongroup: {
		child: {
			required: {
				button: listItemTags.button
			}
		}
	},
	horizontallist: {
		attrs: {
			selected: {
				required: true,
				canModify: true
			}
		},
		child: {
			required: {
				item: listItemTags.item
			}
		}
	},
	handlerlist: {
		child: {
			switchitem: listItemTags.switchitem,
			detailitem: listItemTags.detailitem,
			stepperitem: listItemTags.stepperitem,
			menuitem: listItemTags.menuitem
		}
	},
	waypointhandler: {
		child: {
			required: {
				item: listItemTags.item
			}
		}
	},
	verticallist: {
		attrs: {
			selected: {
				required: true,
				canModify: true
			}
		},
		child: {
			item: listItemTags.item
		}
	}
};

var nativeTags = Object.assign({}, normalTags, listTags, listItemTags);

/**
 * 判断是否是 mesh 允许的 tag
 * 
 * @export
 * @param {any} tag 
 * @returns 
 */
function isNativeTag(tag) {
	if (nativeTags[tag]) {
		return true;
	}
	return false;
}

/**
 * 判断是不是列表类型的 tag
 * 
 * @export
 * @param {any} tag 
 * @returns 
 */
function isList(tag) {
	return !isUndefined(listTags[tag]);
}

/**
 * 判断是不是列表子组件 tag
 * 
 * @export
 * @param {any} node 
 * @returns 
 */
function isListItem(node) {
	var tag = node.sel || node.nodeName.toLowerCase();
	return !isUndefined(listItemTags[tag]) || !isNullOrUndef(node.parentModuleId) && !isNullOrUndef(node.index);
}

/**
 * 检查 patch 的 input 中的 tag 是否合法
 * 由于没有精力维护，现在只是在 mesh-envoy-tag 去做校验，已经取消了这个校验
 * 
 * @export
 * @param {any} input 
 */

function createElement(tagName) {
	return document.createElement(tagName);
}

function createElementNS(namespaceURI, qualifiedName) {
	return document.createElementNS(namespaceURI, qualifiedName);
}

function createTextNode(text) {
	return document.createTextNode(text);
}

function createComment(text) {
	return document.createComment(text);
}

function insertBefore(parentNode, newNode, referenceNode) {
	parentNode.insertBefore(newNode, referenceNode);
	patchQueue.push({
		type: 'insert',
		parent: parentNode.moduleId,
		referenceNode: referenceNode && referenceNode.moduleId,
		node: stripDom(newNode)
	});
}

function removeChild(node, child) {
	patchQueue.push({
		type: 'remove',
		parent: node.moduleId,
		node: stripDom(child)
	});
	node.removeChild(child);
}

function appendChild(node, child) {
	node.appendChild(child);
}

function parentNode(node) {
	return node.parentNode;
}

function nextSibling(node) {
	return node.nextSibling;
}

function tagName(elm) {
	return elm.tagName;
}

function setTextContent(node, text) {
	if (isListItem(node.parentNode)) {
		patchQueue.push({
			type: 'text',
			moduleId: node.parentNode.parentModuleId,
			index: node.parentNode.index,
			text: text.toString()
		});
	} else {
		patchQueue.push({
			type: 'text',
			moduleId: node.parentNode.moduleId,
			text: text.toString()
		});
	}
	node.textContent = text;
}

function getTextContent(node) {
	return node.textContent;
}

function isElement(node) {
	return node.nodeType === 1;
}

function isText(node) {
	return node.nodeType === 3;
}

function isComment(node) {
	return node.nodeType === 8;
}

var domApi = {
	createElement: createElement,
	createElementNS: createElementNS,
	createTextNode: createTextNode,
	createComment: createComment,
	insertBefore: insertBefore,
	removeChild: removeChild,
	appendChild: appendChild,
	parentNode: parentNode,
	nextSibling: nextSibling,
	tagName: tagName,
	setTextContent: setTextContent,
	getTextContent: getTextContent,
	isElement: isElement,
	isText: isText,
	isComment: isComment
};

var commonjsGlobal = typeof window !== 'undefined' ? window : typeof global !== 'undefined' ? global : typeof self !== 'undefined' ? self : {};

function commonjsRequire () {
	throw new Error('Dynamic requires are not currently supported by rollup-plugin-commonjs');
}

function unwrapExports (x) {
	return x && x.__esModule ? x['default'] : x;
}

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

var vnode_1$1 = createCommonjsModule(function (module, exports) {
"use strict";
function vnode(sel, data, children, text, elm) {
    var key = data === undefined ? undefined : data.key;
    return { sel: sel, data: data, children: children,
        text: text, elm: elm, key: key };
}
exports.vnode = vnode;
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = vnode;

});

var array = Array.isArray;
function primitive(s) {
    return typeof s === 'string' || typeof s === 'number';
}
var primitive_1 = primitive;


var is$1 = {
	array: array,
	primitive: primitive_1
};

var htmldomapi = createCommonjsModule(function (module, exports) {
"use strict";
function createElement(tagName) {
    return document.createElement(tagName);
}
function createElementNS(namespaceURI, qualifiedName) {
    return document.createElementNS(namespaceURI, qualifiedName);
}
function createTextNode(text) {
    return document.createTextNode(text);
}
function createComment(text) {
    return document.createComment(text);
}
function insertBefore(parentNode, newNode, referenceNode) {
    parentNode.insertBefore(newNode, referenceNode);
}
function removeChild(node, child) {
    node.removeChild(child);
}
function appendChild(node, child) {
    node.appendChild(child);
}
function parentNode(node) {
    return node.parentNode;
}
function nextSibling(node) {
    return node.nextSibling;
}
function tagName(elm) {
    return elm.tagName;
}
function setTextContent(node, text) {
    node.textContent = text;
}
function getTextContent(node) {
    return node.textContent;
}
function isElement(node) {
    return node.nodeType === 1;
}
function isText(node) {
    return node.nodeType === 3;
}
function isComment(node) {
    return node.nodeType === 8;
}
exports.htmlDomApi = {
    createElement: createElement,
    createElementNS: createElementNS,
    createTextNode: createTextNode,
    createComment: createComment,
    insertBefore: insertBefore,
    removeChild: removeChild,
    appendChild: appendChild,
    parentNode: parentNode,
    nextSibling: nextSibling,
    tagName: tagName,
    setTextContent: setTextContent,
    getTextContent: getTextContent,
    isElement: isElement,
    isText: isText,
    isComment: isComment,
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = exports.htmlDomApi;

});

var h_1$1 = createCommonjsModule(function (module, exports) {
"use strict";
var vnode_1 = vnode_1$1;
var is = is$1;
function addNS(data, children, sel) {
    data.ns = 'http://www.w3.org/2000/svg';
    if (sel !== 'foreignObject' && children !== undefined) {
        for (var i = 0; i < children.length; ++i) {
            var childData = children[i].data;
            if (childData !== undefined) {
                addNS(childData, children[i].children, children[i].sel);
            }
        }
    }
}
function h(sel, b, c) {
    var data = {}, children, text, i;
    if (c !== undefined) {
        data = b;
        if (is.array(c)) {
            children = c;
        }
        else if (is.primitive(c)) {
            text = c;
        }
        else if (c && c.sel) {
            children = [c];
        }
    }
    else if (b !== undefined) {
        if (is.array(b)) {
            children = b;
        }
        else if (is.primitive(b)) {
            text = b;
        }
        else if (b && b.sel) {
            children = [b];
        }
        else {
            data = b;
        }
    }
    if (is.array(children)) {
        for (i = 0; i < children.length; ++i) {
            if (is.primitive(children[i]))
                children[i] = vnode_1.vnode(undefined, undefined, undefined, children[i]);
        }
    }
    if (sel[0] === 's' && sel[1] === 'v' && sel[2] === 'g' &&
        (sel.length === 3 || sel[3] === '.' || sel[3] === '#')) {
        addNS(data, children, sel);
    }
    return vnode_1.vnode(sel, data, children, text, undefined);
}
exports.h = h;

Object.defineProperty(exports, "__esModule", { value: true });
exports.default = h;

});

var h$1 = unwrapExports(h_1$1);

var thunk$1 = createCommonjsModule(function (module, exports) {
"use strict";
var h_1 = h_1$1;
function copyToThunk(vnode, thunk) {
    thunk.elm = vnode.elm;
    vnode.data.fn = thunk.data.fn;
    vnode.data.args = thunk.data.args;
    thunk.data = vnode.data;
    thunk.children = vnode.children;
    thunk.text = vnode.text;
    thunk.elm = vnode.elm;
}
function init(thunk) {
    var cur = thunk.data;
    var vnode = cur.fn.apply(undefined, cur.args);
    copyToThunk(vnode, thunk);
}
function prepatch(oldVnode, thunk) {
    var i, old = oldVnode.data, cur = thunk.data;
    var oldArgs = old.args, args = cur.args;
    if (old.fn !== cur.fn || oldArgs.length !== args.length) {
        copyToThunk(cur.fn.apply(undefined, args), thunk);
    }
    for (i = 0; i < args.length; ++i) {
        if (oldArgs[i] !== args[i]) {
            copyToThunk(cur.fn.apply(undefined, args), thunk);
            return;
        }
    }
    copyToThunk(oldVnode, thunk);
}
exports.thunk = function thunk(sel, key, fn, args) {
    if (args === undefined) {
        args = fn;
        fn = key;
        key = undefined;
    }
    return h_1.h(sel, {
        key: key,
        hook: { init: init, prepatch: prepatch },
        fn: fn,
        args: args
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = exports.thunk;

});

var vnode_1 = vnode_1$1;
var is = is$1;
var htmldomapi_1 = htmldomapi;
function isUndef(s) { return s === undefined; }
function isDef(s) { return s !== undefined; }
var emptyNode = vnode_1.default('', {}, [], undefined, undefined);
function sameVnode(vnode1, vnode2) {
    return vnode1.key === vnode2.key && vnode1.sel === vnode2.sel;
}
function isVnode(vnode) {
    return vnode.sel !== undefined;
}
function createKeyToOldIdx(children, beginIdx, endIdx) {
    var i, map = {}, key, ch;
    for (i = beginIdx; i <= endIdx; ++i) {
        ch = children[i];
        if (ch != null) {
            key = ch.key;
            if (key !== undefined)
                map[key] = i;
        }
    }
    return map;
}
var hooks = ['create', 'update', 'remove', 'destroy', 'pre', 'post'];
var h_1 = h_1$1;
var h = h_1.h;
var thunk_1 = thunk$1;
var thunk = thunk_1.thunk;
function init(modules, domApi) {
    var i, j, cbs = {};
    var api = domApi !== undefined ? domApi : htmldomapi_1.default;
    for (i = 0; i < hooks.length; ++i) {
        cbs[hooks[i]] = [];
        for (j = 0; j < modules.length; ++j) {
            var hook = modules[j][hooks[i]];
            if (hook !== undefined) {
                cbs[hooks[i]].push(hook);
            }
        }
    }
    function emptyNodeAt(elm) {
        var id = elm.id ? '#' + elm.id : '';
        var c = elm.className ? '.' + elm.className.split(' ').join('.') : '';
        return vnode_1.default(api.tagName(elm).toLowerCase() + id + c, {}, [], undefined, elm);
    }
    function createRmCb(childElm, listeners) {
        return function rmCb() {
            if (--listeners === 0) {
                var parent_1 = api.parentNode(childElm);
                api.removeChild(parent_1, childElm);
            }
        };
    }
    function createElm(vnode, insertedVnodeQueue) {
        var i, data = vnode.data;
        if (data !== undefined) {
            if (isDef(i = data.hook) && isDef(i = i.init)) {
                i(vnode);
                data = vnode.data;
            }
        }
        var children = vnode.children, sel = vnode.sel;
        if (sel === '!') {
            if (isUndef(vnode.text)) {
                vnode.text = '';
            }
            vnode.elm = api.createComment(vnode.text);
        }
        else if (sel !== undefined) {
            // Parse selector
            var hashIdx = sel.indexOf('#');
            var dotIdx = sel.indexOf('.', hashIdx);
            var hash = hashIdx > 0 ? hashIdx : sel.length;
            var dot = dotIdx > 0 ? dotIdx : sel.length;
            var tag = hashIdx !== -1 || dotIdx !== -1 ? sel.slice(0, Math.min(hash, dot)) : sel;
            var elm = vnode.elm = isDef(data) && isDef(i = data.ns) ? api.createElementNS(i, tag)
                : api.createElement(tag);
            if (hash < dot)
                elm.id = sel.slice(hash + 1, dot);
            if (dotIdx > 0)
                elm.className = sel.slice(dot + 1).replace(/\./g, ' ');
            for (i = 0; i < cbs.create.length; ++i)
                cbs.create[i](emptyNode, vnode);
            if (is.array(children)) {
                for (i = 0; i < children.length; ++i) {
                    var ch = children[i];
                    if (ch != null) {
                        api.appendChild(elm, createElm(ch, insertedVnodeQueue));
                    }
                }
            }
            else if (is.primitive(vnode.text)) {
                api.appendChild(elm, api.createTextNode(vnode.text));
            }
            i = vnode.data.hook; // Reuse variable
            if (isDef(i)) {
                if (i.create)
                    i.create(emptyNode, vnode);
                if (i.insert)
                    insertedVnodeQueue.push(vnode);
            }
        }
        else {
            vnode.elm = api.createTextNode(vnode.text);
        }
        return vnode.elm;
    }
    function addVnodes(parentElm, before, vnodes, startIdx, endIdx, insertedVnodeQueue) {
        for (; startIdx <= endIdx; ++startIdx) {
            var ch = vnodes[startIdx];
            if (ch != null) {
                api.insertBefore(parentElm, createElm(ch, insertedVnodeQueue), before);
            }
        }
    }
    function invokeDestroyHook(vnode) {
        var i, j, data = vnode.data;
        if (data !== undefined) {
            if (isDef(i = data.hook) && isDef(i = i.destroy))
                i(vnode);
            for (i = 0; i < cbs.destroy.length; ++i)
                cbs.destroy[i](vnode);
            if (vnode.children !== undefined) {
                for (j = 0; j < vnode.children.length; ++j) {
                    i = vnode.children[j];
                    if (i != null && typeof i !== "string") {
                        invokeDestroyHook(i);
                    }
                }
            }
        }
    }
    function removeVnodes(parentElm, vnodes, startIdx, endIdx) {
        for (; startIdx <= endIdx; ++startIdx) {
            var i_1 = void 0, listeners = void 0, rm = void 0, ch = vnodes[startIdx];
            if (ch != null) {
                if (isDef(ch.sel)) {
                    invokeDestroyHook(ch);
                    listeners = cbs.remove.length + 1;
                    rm = createRmCb(ch.elm, listeners);
                    for (i_1 = 0; i_1 < cbs.remove.length; ++i_1)
                        cbs.remove[i_1](ch, rm);
                    if (isDef(i_1 = ch.data) && isDef(i_1 = i_1.hook) && isDef(i_1 = i_1.remove)) {
                        i_1(ch, rm);
                    }
                    else {
                        rm();
                    }
                }
                else {
                    api.removeChild(parentElm, ch.elm);
                }
            }
        }
    }
    function updateChildren(parentElm, oldCh, newCh, insertedVnodeQueue) {
        var oldStartIdx = 0, newStartIdx = 0;
        var oldEndIdx = oldCh.length - 1;
        var oldStartVnode = oldCh[0];
        var oldEndVnode = oldCh[oldEndIdx];
        var newEndIdx = newCh.length - 1;
        var newStartVnode = newCh[0];
        var newEndVnode = newCh[newEndIdx];
        var oldKeyToIdx;
        var idxInOld;
        var elmToMove;
        var before;
        while (oldStartIdx <= oldEndIdx && newStartIdx <= newEndIdx) {
            if (oldStartVnode == null) {
                oldStartVnode = oldCh[++oldStartIdx]; // Vnode might have been moved left
            }
            else if (oldEndVnode == null) {
                oldEndVnode = oldCh[--oldEndIdx];
            }
            else if (newStartVnode == null) {
                newStartVnode = newCh[++newStartIdx];
            }
            else if (newEndVnode == null) {
                newEndVnode = newCh[--newEndIdx];
            }
            else if (sameVnode(oldStartVnode, newStartVnode)) {
                patchVnode(oldStartVnode, newStartVnode, insertedVnodeQueue);
                oldStartVnode = oldCh[++oldStartIdx];
                newStartVnode = newCh[++newStartIdx];
            }
            else if (sameVnode(oldEndVnode, newEndVnode)) {
                patchVnode(oldEndVnode, newEndVnode, insertedVnodeQueue);
                oldEndVnode = oldCh[--oldEndIdx];
                newEndVnode = newCh[--newEndIdx];
            }
            else if (sameVnode(oldStartVnode, newEndVnode)) {
                patchVnode(oldStartVnode, newEndVnode, insertedVnodeQueue);
                api.insertBefore(parentElm, oldStartVnode.elm, api.nextSibling(oldEndVnode.elm));
                oldStartVnode = oldCh[++oldStartIdx];
                newEndVnode = newCh[--newEndIdx];
            }
            else if (sameVnode(oldEndVnode, newStartVnode)) {
                patchVnode(oldEndVnode, newStartVnode, insertedVnodeQueue);
                api.insertBefore(parentElm, oldEndVnode.elm, oldStartVnode.elm);
                oldEndVnode = oldCh[--oldEndIdx];
                newStartVnode = newCh[++newStartIdx];
            }
            else {
                if (oldKeyToIdx === undefined) {
                    oldKeyToIdx = createKeyToOldIdx(oldCh, oldStartIdx, oldEndIdx);
                }
                idxInOld = oldKeyToIdx[newStartVnode.key];
                if (isUndef(idxInOld)) {
                    api.insertBefore(parentElm, createElm(newStartVnode, insertedVnodeQueue), oldStartVnode.elm);
                    newStartVnode = newCh[++newStartIdx];
                }
                else {
                    elmToMove = oldCh[idxInOld];
                    if (elmToMove.sel !== newStartVnode.sel) {
                        api.insertBefore(parentElm, createElm(newStartVnode, insertedVnodeQueue), oldStartVnode.elm);
                    }
                    else {
                        patchVnode(elmToMove, newStartVnode, insertedVnodeQueue);
                        oldCh[idxInOld] = undefined;
                        api.insertBefore(parentElm, elmToMove.elm, oldStartVnode.elm);
                    }
                    newStartVnode = newCh[++newStartIdx];
                }
            }
        }
        if (oldStartIdx > oldEndIdx) {
            before = newCh[newEndIdx + 1] == null ? null : newCh[newEndIdx + 1].elm;
            addVnodes(parentElm, before, newCh, newStartIdx, newEndIdx, insertedVnodeQueue);
        }
        else if (newStartIdx > newEndIdx) {
            removeVnodes(parentElm, oldCh, oldStartIdx, oldEndIdx);
        }
    }
    function patchVnode(oldVnode, vnode, insertedVnodeQueue) {
        var i, hook;
        if (isDef(i = vnode.data) && isDef(hook = i.hook) && isDef(i = hook.prepatch)) {
            i(oldVnode, vnode);
        }
        var elm = vnode.elm = oldVnode.elm;
        var oldCh = oldVnode.children;
        var ch = vnode.children;
        if (oldVnode === vnode)
            return;
        if (vnode.data !== undefined) {
            for (i = 0; i < cbs.update.length; ++i)
                cbs.update[i](oldVnode, vnode);
            i = vnode.data.hook;
            if (isDef(i) && isDef(i = i.update))
                i(oldVnode, vnode);
        }
        if (isUndef(vnode.text)) {
            if (isDef(oldCh) && isDef(ch)) {
                if (oldCh !== ch)
                    updateChildren(elm, oldCh, ch, insertedVnodeQueue);
            }
            else if (isDef(ch)) {
                if (isDef(oldVnode.text))
                    api.setTextContent(elm, '');
                addVnodes(elm, null, ch, 0, ch.length - 1, insertedVnodeQueue);
            }
            else if (isDef(oldCh)) {
                removeVnodes(elm, oldCh, 0, oldCh.length - 1);
            }
            else if (isDef(oldVnode.text)) {
                api.setTextContent(elm, '');
            }
        }
        else if (oldVnode.text !== vnode.text) {
            api.setTextContent(elm, vnode.text);
        }
        if (isDef(hook) && isDef(i = hook.postpatch)) {
            i(oldVnode, vnode);
        }
    }
    return function patch(oldVnode, vnode) {
        var i, elm, parent;
        var insertedVnodeQueue = [];
        for (i = 0; i < cbs.pre.length; ++i)
            cbs.pre[i]();
        if (!isVnode(oldVnode)) {
            oldVnode = emptyNodeAt(oldVnode);
        }
        if (sameVnode(oldVnode, vnode)) {
            patchVnode(oldVnode, vnode, insertedVnodeQueue);
        }
        else {
            elm = oldVnode.elm;
            parent = api.parentNode(elm);
            createElm(vnode, insertedVnodeQueue);
            if (parent !== null) {
                api.insertBefore(parent, vnode.elm, api.nextSibling(elm));
                removeVnodes(parent, [oldVnode], 0, 0);
            }
        }
        for (i = 0; i < insertedVnodeQueue.length; ++i) {
            insertedVnodeQueue[i].data.hook.insert(insertedVnodeQueue[i]);
        }
        for (i = 0; i < cbs.post.length; ++i)
            cbs.post[i]();
        return vnode;
    };
}
var init_1 = init;


var snabbdom = {
	h: h,
	thunk: thunk,
	init: init_1
};

var eventlisteners = createCommonjsModule(function (module, exports) {
"use strict";
function invokeHandler(handler, vnode, event) {
    if (typeof handler === "function") {
        // call function handler
        handler.call(vnode, event, vnode);
    }
    else if (typeof handler === "object") {
        // call handler with arguments
        if (typeof handler[0] === "function") {
            // special case for single argument for performance
            if (handler.length === 2) {
                handler[0].call(vnode, handler[1], event, vnode);
            }
            else {
                var args = handler.slice(1);
                args.push(event);
                args.push(vnode);
                handler[0].apply(vnode, args);
            }
        }
        else {
            // call multiple handlers
            for (var i = 0; i < handler.length; i++) {
                invokeHandler(handler[i]);
            }
        }
    }
}
function handleEvent(event, vnode) {
    var name = event.type, on = vnode.data.on;
    // call event handler(s) if exists
    if (on && on[name]) {
        invokeHandler(on[name], vnode, event);
    }
}
function createListener() {
    return function handler(event) {
        handleEvent(event, handler.vnode);
    };
}
function updateEventListeners(oldVnode, vnode) {
    var oldOn = oldVnode.data.on, oldListener = oldVnode.listener, oldElm = oldVnode.elm, on = vnode && vnode.data.on, elm = (vnode && vnode.elm), name;
    // optimization for reused immutable handlers
    if (oldOn === on) {
        return;
    }
    // remove existing listeners which no longer used
    if (oldOn && oldListener) {
        // if element changed or deleted we remove all existing listeners unconditionally
        if (!on) {
            for (name in oldOn) {
                // remove listener if element was changed or existing listeners removed
                oldElm.removeEventListener(name, oldListener, false);
            }
        }
        else {
            for (name in oldOn) {
                // remove listener if existing listener removed
                if (!on[name]) {
                    oldElm.removeEventListener(name, oldListener, false);
                }
            }
        }
    }
    // add new listeners which has not already attached
    if (on) {
        // reuse existing listener or create new
        var listener = vnode.listener = oldVnode.listener || createListener();
        // update vnode for listener
        listener.vnode = vnode;
        // if element changed or added we add all needed listeners unconditionally
        if (!oldOn) {
            for (name in on) {
                // add listener if element was changed or new listeners added
                elm.addEventListener(name, listener, false);
            }
        }
        else {
            for (name in on) {
                // add listener if new listener added
                if (!oldOn[name]) {
                    elm.addEventListener(name, listener, false);
                }
            }
        }
    }
}
exports.eventListenersModule = {
    create: updateEventListeners,
    update: updateEventListeners,
    destroy: updateEventListeners
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = exports.eventListenersModule;

});

var eventlisteners$1 = unwrapExports(eventlisteners);

var attributes = createCommonjsModule(function (module, exports) {
"use strict";
var NamespaceURIs = {
    "xlink": "http://www.w3.org/1999/xlink"
};
var booleanAttrs = ["allowfullscreen", "async", "autofocus", "autoplay", "checked", "compact", "controls", "declare",
    "default", "defaultchecked", "defaultmuted", "defaultselected", "defer", "disabled", "draggable",
    "enabled", "formnovalidate", "hidden", "indeterminate", "inert", "ismap", "itemscope", "loop", "multiple",
    "muted", "nohref", "noresize", "noshade", "novalidate", "nowrap", "open", "pauseonexit", "readonly",
    "required", "reversed", "scoped", "seamless", "selected", "sortable", "spellcheck", "translate",
    "truespeed", "typemustmatch", "visible"];
var booleanAttrsDict = Object.create(null);
for (var i = 0, len = booleanAttrs.length; i < len; i++) {
    booleanAttrsDict[booleanAttrs[i]] = true;
}
function updateAttrs(oldVnode, vnode) {
    var key, elm = vnode.elm, oldAttrs = oldVnode.data.attrs, attrs = vnode.data.attrs, namespaceSplit;
    if (!oldAttrs && !attrs)
        return;
    if (oldAttrs === attrs)
        return;
    oldAttrs = oldAttrs || {};
    attrs = attrs || {};
    // update modified attributes, add new attributes
    for (key in attrs) {
        var cur = attrs[key];
        var old = oldAttrs[key];
        if (old !== cur) {
            if (booleanAttrsDict[key]) {
                if (cur) {
                    elm.setAttribute(key, "");
                }
                else {
                    elm.removeAttribute(key);
                }
            }
            else {
                namespaceSplit = key.split(":");
                if (namespaceSplit.length > 1 && NamespaceURIs.hasOwnProperty(namespaceSplit[0])) {
                    elm.setAttributeNS(NamespaceURIs[namespaceSplit[0]], key, cur);
                }
                else {
                    elm.setAttribute(key, cur);
                }
            }
        }
    }
    // remove removed attributes
    // use `in` operator since the previous `for` iteration uses it (.i.e. add even attributes with undefined value)
    // the other option is to remove all attributes with value == undefined
    for (key in oldAttrs) {
        if (!(key in attrs)) {
            elm.removeAttribute(key);
        }
    }
}
exports.attributesModule = { create: updateAttrs, update: updateAttrs };
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = exports.attributesModule;

});

var attributes$1 = unwrapExports(attributes);

var instance = {
	/**
  * snabbdom create hook
  * 
  * @param {any} oldVnode 
  * @param {any} newVnode 
  */
	create: function create(oldVnode, newVnode) {
		var elm = newVnode.elm;
		// 添加 moduleId
		newVnode.moduleId = moduleId++;
		if (isList(newVnode.sel)) {
			var children = newVnode.children;
			// 由于客户端那边对于列表对象的操作都需要传入父 moduleId 和 index 来进行操作, 所以子组件都标记了父组件的 moduleId , 所以不管嵌套了多少层子对象，要操作它都必须在顶层的 list 那里进行操作， 所以多层嵌套依然父对象为 list 那个对象
			children && children.forEach(function (item, index) {
				if (isNullOrUndef(item)) {
					return;
				}
				item.parentModuleId = newVnode.moduleId;
				item.index = index;
				if (item.children) {
					item.children.forEach(function (childItem) {
						if (isNullOrUndef(childItem) || isNullOrUndef(childItem.sel)) {
							return;
						}
						childItem.index = index;
						childItem.parentModuleId = newVnode.moduleId;
					});
				}
			});
		} else if (isListItem(newVnode)) {
			elm.index = newVnode.index;
			elm.parentModuleId = newVnode.parentModuleId;
		}

		elm.moduleId = newVnode.moduleId;
		elm.attrs = newVnode.data.attrs || {};
	},

	/**
  * update hook
  * 
  * @param {any} oldVnode 
  * @param {any} vnode 
  */
	update: function update(oldVnode, vnode) {
		var key = void 0,
		    cur = void 0,
		    old = void 0,
		    elm = vnode.elm,
		    change = [],
		    moduleId = oldVnode.moduleId,
		    oldAttrs = oldVnode.data.attrs,
		    attrs = vnode.data.attrs;

		if (!isNullOrUndef(oldVnode.moduleId)) {
			vnode.moduleId = oldVnode.moduleId;
			vnode.instance = oldVnode.instance;
			vnode.data.hook = oldVnode.data.hook;
			vnode.vNode = oldVnode.vNode;
			if (!isNullOrUndef(oldVnode.parentModuleId)) {
				vnode.parentModuleId = oldVnode.parentModuleId;
				vnode.index = oldVnode.index;
			}
		}

		if (!oldAttrs && !attrs) return;
		if (oldAttrs === attrs) return;
		oldAttrs = oldAttrs || {};
		attrs = attrs || {};

		// 这里对新旧两个 attr 进行了 diff，来更新 vnode 对象

		// 旧 attr 有，新的没有，需要移除 key
		for (key in oldAttrs) {
			if (isNullOrUndef(attrs[key])) {
				change.push({
					type: 'remove',
					key: key
				});
				delete elm[key];
			}
		}

		// 对比两个值是否一样，标记修改
		for (key in attrs) {
			cur = attrs[key];
			old = oldAttrs[key];
			if (old !== cur && (key !== 'value' || elm[key] !== cur)) {
				change.push({
					type: 'modify',
					key: key,
					value: cur
				});
				elm[key] = cur;
			}
		}

		// 发送修改，区分 listItem，传入创建时的 parentModuleId 和 index 索引
		if (change.length > 0) {
			if (isListItem(oldVnode)) {
				patchQueue.push({
					type: 'props',
					moduleId: oldVnode.parentModuleId,
					index: oldVnode.index,
					change: change
				});
			} else {
				patchQueue.push({
					type: 'props',
					moduleId: moduleId,
					change: change
				});
			}
		}
	}
};

var transform = {
	create: function create(oldVnode, newVnode) {
		var children = newVnode.children;
		// MARK: force the text node to be a String 
		// because when text be 0 
		// document.createText didn't create it
		if (!isUndefined(children) && children.length === 1 && !isUndefined(children[0]) && isUndefined(children[0].sel)) {
			if (!isObject(children[0])) {
				throwError(newVnode.sel + ' has unvalid children ' + children[0]);
			} else if (isNumber(children[0].text)) {
				children[0].text = '' + children[0].text;
			}
		}

		var attrs = newVnode.data.attrs;
		for (var key in attrs) {
			if (isNullOrUndef(attrs[key])) {
				delete attrs[key];
			}
			if (attrs[key] && attrs[key].constructor.name === 'ObservableArray') {
				attrs[key] = attrs[key].peek();
			}
		}
	}
};

/**
 * This module is to binding vnode reference to the instance ref function
 */
function bindRef(vnode) {
	if (!vnode) {
		return;
	}
	// 查找 vnode 的 attr 是否有 refs 这个 key 并且是一个函数
	// https://react-cn.github.io/react/docs/more-about-refs.html
	var attrs = vnode.data.attrs;
	if (attrs && isFunction(attrs.refs)) {
		attrs.refs(vnode);
	}
}

var refs = {
	/**
  * create hook
  * 
  * @param {any} oldVnode 
  * @param {any} newVnode 
  */
	create: function create(oldVnode, newVnode) {
		bindRef(newVnode);
		var children = newVnode.children;
		// MARK: Map children and children isn't a text node 
		if (!isUndefined(children) && children.length >= 1 && !isUndefined(children[0]) && !isUndefined(children[0].sel)) {
			children.map(function (item) {
				bindRef(item);
			});
		}
	}
};

var patch = snabbdom.init([eventlisteners$1, attributes$1, instance, transform, refs], domApi);

function createVNode() {
	return h$1.apply(undefined, arguments);
}

/**
 * 页面路由
 * 需要注意的是，如果直接通过 mesh.back 方法返回页面，将不会触发 route 的退栈,
 * 此时栈的队列会出错
 */

var Route = function () {
	function Route() {
		classCallCheck(this, Route);
	}

	createClass(Route, null, [{
		key: 'push',

		/**
   * push 一个新的页面
   * 
   * @static
   * @param {any} input 
   * @param {any} data 
   * @memberof Route
   */
		value: function push(input, data) {
			var stack = void 0;
			// 可以 push 一个 Component 也可以直接 push jsx tag
			if (isFunction(input)) {
				stack = render(createVNode(input, {}), 'push', data);
			} else {
				stack = render(input, 'push', data);
			}
			// 对 handlerPage 做特殊标记，用于 FeatureManager _registMissionFeatureControls 判断是否是 handler
			if (stack.sel === 'handlerpage') {
				stack.isHandlerPage = true;
			} else if (stack.sel === 'mainpage') {
				stack.isMainPage = true;
			}
			// push 该页面进到栈中，方便索引
			this.stack.push(stack);
		}
		/**
   * present 一个 view, 逻辑是一样的，只不过调用的参数不同
   * 
   * @static
   * @param {any} input 
   * @memberof Route
   */

	}, {
		key: 'present',
		value: function present(input) {
			var stack = void 0;
			if (isFunction(input)) {
				stack = render(createVNode(input, {}), 'present');
			} else {
				stack = render(input, 'present');
			}
			if (stack.sel === 'handlerpage') {
				stack.isHandlerPage = true;
			} else if (stack.sel === 'mainpage') {
				stack.isMainPage = true;
			}
			this.stack.push(stack);
		}

		/**
   * 返回上一个页面
   * 
   * @static
   * @memberof Route
   */

	}, {
		key: 'back',
		value: function back() {
			if (typeof mesh !== 'undefined') {
				// 页面退栈
				this.stack.pop();
				mesh.back();
			}
		}

		/**
   * 回到首页
   * 
   * @static
   * @memberof Route
   */

	}, {
		key: 'backToRoot',
		value: function backToRoot() {
			if (typeof mesh !== 'undefined') {
				// 将栈清空
				this.stack = [];
				mesh.backToRoot();
			}
		}
	}]);
	return Route;
}();

Route.stack = [];

/**
 *  废弃，由于现在已经不能直接在 package 中调用 feature，所以这个 testRender 已经无法实现自动化测试
 * @param {} input 
 */
function testRender(input) {
	if (input === NO_OP) {
		return;
	}

	var patchResult = createPatch(document.createElement('body'), input, {});

	return patchResult;
}

/**
 *  核心 render 方法，创建一个新的 view
 * @param {*} input 
 * @param {'push'|'present'} type 
 * @param {*} data 
 */
function render(input) {
	var type = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 'push';
	var data = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

	if (input === NO_OP) {
		return;
	}

	document.body.innerHTML = "";
	var patchResult = createPatch(document.createElement('body'), input, {});

	document.body.appendChild(patchResult.elm);
	sendDom(patchResult.elm, type, data);
	return patchResult;
}

/**
 * diff 方法
 * @param {} Node 
 * @param {} VNode 
 * @param {} context 
 */
function createPatch(Node, VNode, context) {
	//TODO: Check VNode is Component and don't create instance if state,props don't update, just update the ref instance 
	// 判断是否依然是同一个实例，不是的话重新将实例 mount ，是的话直接用新的 props 更新实例
	if (shouldPatch(Node.instance, VNode)) {
		var input = mount(VNode, context);
		var patchResult = patch(Node, input);
		createModuleMap(patchResult);
		return patchResult;
	} else {
		return applyProps(Node.instance, Object.assign({}, {
			children: VNode.children
		}, VNode.data.attrs), context);
	}
}

/**
 * 判断是否是一个实例
 * @param {} vNode 
 * @param {} newVNode 
 */
function shouldPatch(vNode, newVNode) {
	if (!vNode || vNode.constructor.name !== newVNode.sel.name) {
		return true;
	}
	return false;
}

/**
 *  实例更新方法
 * @param {} component 
 * @param {} nextProps 
 * @param {} context 
 */
function applyProps(component, nextProps, context) {
	if (!component._blockRender && !component._unmounted) {
		var state = component.state;
		var props = component.props;

		var nextInput = component._undateComponent(state, state, props, nextProps, context, false, false);
		var didUpdate = true;

		if (nextInput === NO_OP) {
			// 没有更新
			nextInput = component._lastInput;
			didUpdate = false;
		} else if (isStringOrNumber(nextInput)) {
			// 文本节点
			nextInput = {
				text: nextInput.toString()
			};
		} else if (isArray(nextInput)) {
			// render 方法必须有一个顶层节点
			if (process.env.NODE_ENV !== 'production') {
				throwError('a valid Envoy VNode must be returned from a component render. You may have returned an array or an invalid object');
			}
			throwError();
		}

		var lastInput = component._lastInput;

		if (didUpdate) {
			// 可以通过 subLifecycle 注入一些更新操作，但其实并未使用
			var subLifecycle = component._lifecycle;

			if (!subLifecycle) {
				subLifecycle = new Lifecycle();
			} else {
				subLifecycle.listeners = [];
			}
			component._lifecycle = subLifecycle;
			// context 主要是用于整个组件树共享数据，比如 mobx 的 store
			var childContext = component.getChildContext();

			if (isNullOrUndef(childContext)) {
				childContext = context;
			} else {
				childContext = Object.assign({}, context, component._childContext, childContext);
			}

			// component 的 patch 方法实际指向了 createPatch 方法
			component._lastInput = component._patch(lastInput, nextInput, childContext);
			subLifecycle.trigger();
			// 执行 patch 队列
			component.excutePatchQueue();
			// 触发生命周期
			component.componentDidUpdate(nextProps, state);
		}
	} else {
		component.props = nextProps;
	}

	return component._lastInput;
}

/**
 * 递归的将 vNode mount
 * @param {} input 
 * @param {} context 
 * @param {} _parent 
 */
function mount(input, context, _parent) {
	if (isFunction(input.sel)) {
		// 一个 Component ，生成实例
		// 标记当前页面的 root 的实例, 方便子组件可以直接索引到根组件
		var _isRoot = !_parent;
		// 创建实例
		var instance = createClassComponentInstance(input.sel, input.data, context, input.children, input, _isRoot);
		// 父子组件相互索引，根组件索引, 这个实现方式比较的丑，但是在很多时候需要做一些很奇怪的需求的时候有用, 这里可能会有内存问题，不过组件不算多，其实也还好
		if (_parent) {
			instance._parent = _parent;
			if (!Array.isArray(instance._parent._child)) {
				instance._parent._child = [];
			}
			instance._parent._child.push(instance);
			if (instance._parent._isRoot) {
				instance._root = _parent;
			} else {
				Object.defineProperty(instance, '_root', {
					get: function get() {
						return instance._parent._root;
					}
				});
			}
		}

		return instance._lastInput;
	} else {
		if (isArray(input.children)) {
			// 循环子组件，递归 mount
			input.children = input.children.map(function (child) {
				if (isInvalid(child)) {
					return;
				}
				return mount(child, context, _parent);
			});
		}
		return input;
	}
}

/**
 * 创建实例
 * @param {*} Component 
 * @param {*} data 
 * @param {*} context 
 * @param {*} children 
 * @param {*} vNode 
 * @param {*} isRoot 
 */
function createClassComponentInstance(Component, data, context, children, vNode, isRoot) {
	var instance = new Component(data.attrs, context);
	instance._isRoot = isRoot;
	instance._blockSetState = false;
	instance.props.children = instance._childNodes = children;
	instance.componentWillMount && instance.componentWillMount();
	var input = instance.render(instance.props, instance.context);
	var childContext = instance.getChildContext();
	if (!isNullOrUndef(childContext)) {
		context = Object.assign({}, context, childContext);
	}
	instance.excutePatchQueue = excutePatchQueue;
	instance.route = Route;

	instance._patch = createPatch;
	instance._vNode = input;
	instance._lastInput = mount(input, context, instance);

	if (!instance._lastInput.instance) {
		// 在 node 中索引 实例，很多时候因为 mount 之后在 ref 中拿到的是 node，所以需要索引实例来关联起来
		instance._lastInput.instance = instance;
		// snabbdom 的 hook，当节点被移除是将实例 destroy
		instance._lastInput.data.hook = {
			remove: function remove(node, cb) {
				instance.componentWillUnmount && instance.componentWillUnmount();
				delete MODULE_MAP[node.moduleId];
				instance._unmounted = true;
				cb();
			}
		};
	}

	// 由于是递归的 mount ，所以讲 DidMount 方法放到宏任务中异步执行
	if (isFunction(instance.componentDidMount)) {
		Promise.resolve().then(function () {
			instance.componentDidMount();
		});
	}

	if (instance.route.stack.length === 0 && instance._isRoot) {
		var lastInput = instance._lastInput;
		if (lastInput.sel === 'handlerpage') {
			lastInput.isHandlerPage = true;
		} else if (lastInput.sel === 'mainpage') {
			lastInput.isMainPage = true;
		}
		instance.route.stack.push(lastInput);
	}

	return instance;
}

/**
 *  发送 dom 节点到客户端
 * @param {} dom 
 * @param {} type 
 * @param {} data 
 */
function sendDom(dom, type) {
	var data = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

	if (typeof mesh !== 'undefined') {
		var newDom = stripDom(dom);
		var currentStep = data.currentStep || 1;
		var totalStep = data.totalStep || 1;
		mesh.render(newDom, type, currentStep, totalStep);
	}
}

/**
 * 将每一个 node 根据其 moduleId 存放于 MODULE_MAP 中，用于索引
 * @param {} input 
 */
function createModuleMap(input) {
	if (isNullOrUndef(input)) {
		return;
	}
	if (isNativeTag(input.sel)) {
		MODULE_MAP[input.moduleId] = input;
	} else {
		throwError(input.sel + ' is not supported in envoy ');
	}
	if (isArray(input.children)) {
		// text
		if (input.children.length === 1 && (isNullOrUndef(input.children[0]) || isNullOrUndef(input.children[0].sel))) {
			return;
		}
		input.children.map(function (v) {
			if (isNullOrUndef(v) || isNullOrUndef(v.sel)) {
				return;
			}
			createModuleMap(v);
		});
	}
}

var slice = Array.prototype.slice;

var index$3 = iterativelyWalk;

function iterativelyWalk(nodes, cb) {
    if (!('length' in nodes)) {
        nodes = [nodes];
    }
    
    nodes = slice.call(nodes);

    while(nodes.length) {
        var node = nodes.shift(),
            ret = cb(node);

        if (ret) {
            return ret
        }

        if (node.childNodes && node.childNodes.length) {
            nodes = slice.call(node.childNodes).concat(nodes);
        }
    }
}

var domComment = Comment$1;

function Comment$1(data, owner) {
    if (!(this instanceof Comment$1)) {
        return new Comment$1(data, owner)
    }

    this.data = data;
    this.nodeValue = data;
    this.length = data.length;
    this.ownerDocument = owner || null;
}

Comment$1.prototype.nodeType = 8;
Comment$1.prototype.nodeName = "#comment";

Comment$1.prototype.toString = function _Comment_toString() {
    return "[object Comment]"
};

var domText = DOMText$1;

function DOMText$1(value, owner) {
    if (!(this instanceof DOMText$1)) {
        return new DOMText$1(value)
    }

    this.data = value || "";
    this.length = this.data.length;
    this.ownerDocument = owner || null;
}

DOMText$1.prototype.type = "DOMTextNode";
DOMText$1.prototype.nodeType = 3;
DOMText$1.prototype.nodeName = "#text";

DOMText$1.prototype.toString = function _Text_toString() {
    return this.data
};

DOMText$1.prototype.replaceData = function replaceData(index, length, value) {
    var current = this.data;
    var left = current.substring(0, index);
    var right = current.substring(index + length, current.length);
    this.data = left + value + right;
    this.length = this.data.length;
};

var dispatchEvent_1 = dispatchEvent$2;

function dispatchEvent$2(ev) {
    var elem = this;
    var type = ev.type;

    if (!ev.target) {
        ev.target = elem;
    }

    if (!elem.listeners) {
        elem.listeners = {};
    }

    var listeners = elem.listeners[type];

    if (listeners) {
        return listeners.forEach(function (listener) {
            ev.currentTarget = elem;
            if (typeof listener === 'function') {
                listener(ev);
            } else {
                listener.handleEvent(ev);
            }
        })
    }

    if (elem.parentNode) {
        elem.parentNode.dispatchEvent(ev);
    }
}

var addEventListener_1 = addEventListener$2;

function addEventListener$2(type, listener) {
    var elem = this;

    if (!elem.listeners) {
        elem.listeners = {};
    }

    if (!elem.listeners[type]) {
        elem.listeners[type] = [];
    }

    if (elem.listeners[type].indexOf(listener) === -1) {
        elem.listeners[type].push(listener);
    }
}

var removeEventListener_1 = removeEventListener$2;

function removeEventListener$2(type, listener) {
    var elem = this;

    if (!elem.listeners) {
        return
    }

    if (!elem.listeners[type]) {
        return
    }

    var list = elem.listeners[type];
    var index = list.indexOf(listener);
    if (index !== -1) {
        list.splice(index, 1);
    }
}

var serialize = serializeNode$1;

var voidElements = ["area","base","br","col","embed","hr","img","input","keygen","link","menuitem","meta","param","source","track","wbr"];

function serializeNode$1(node) {
    switch (node.nodeType) {
        case 3:
            return escapeText(node.data)
        case 8:
            return "<!--" + node.data + "-->"
        default:
            return serializeElement(node)
    }
}

function serializeElement(elem) {
    var strings = [];

    var tagname = elem.tagName;

    if (elem.namespaceURI === "http://www.w3.org/1999/xhtml") {
        tagname = tagname.toLowerCase();
    }

    strings.push("<" + tagname + properties(elem) + datasetify(elem));

    if (voidElements.indexOf(tagname) > -1) {
        strings.push(" />");
    } else {
        strings.push(">");

        if (elem.childNodes.length) {
            strings.push.apply(strings, elem.childNodes.map(serializeNode$1));
        } else if (elem.textContent || elem.innerText) {
            strings.push(escapeText(elem.textContent || elem.innerText));
        } else if (elem.innerHTML) {
            strings.push(elem.innerHTML);
        }

        strings.push("</" + tagname + ">");
    }

    return strings.join("")
}

function isProperty(elem, key) {
    var type = typeof elem[key];

    if (key === "style" && Object.keys(elem.style).length > 0) {
      return true
    }

    return elem.hasOwnProperty(key) &&
        (type === "string" || type === "boolean" || type === "number") &&
        key !== "nodeName" && key !== "className" && key !== "tagName" &&
        key !== "textContent" && key !== "innerText" && key !== "namespaceURI" &&  key !== "innerHTML"
}

function stylify(styles) {
    if (typeof styles === 'string') return styles
    var attr = "";
    Object.keys(styles).forEach(function (key) {
        var value = styles[key];
        key = key.replace(/[A-Z]/g, function(c) {
            return "-" + c.toLowerCase();
        });
        attr += key + ":" + value + ";";
    });
    return attr
}

function datasetify(elem) {
    var ds = elem.dataset;
    var props = [];

    for (var key in ds) {
        props.push({ name: "data-" + key, value: ds[key] });
    }

    return props.length ? stringify(props) : ""
}

function stringify(list) {
    var attributes = [];
    list.forEach(function (tuple) {
        var name = tuple.name;
        var value = tuple.value;

        if (name === "style") {
            value = stylify(value);
        }

        attributes.push(name + "=" + "\"" + escapeAttributeValue(value) + "\"");
    });

    return attributes.length ? " " + attributes.join(" ") : ""
}

function properties(elem) {
    var props = [];
    for (var key in elem) {
        if (isProperty(elem, key)) {
            props.push({ name: key, value: elem[key] });
        }
    }

    for (var ns in elem._attributes) {
      for (var attribute in elem._attributes[ns]) {
        var prop = elem._attributes[ns][attribute];
        var name = (prop.prefix ? prop.prefix + ":" : "") + attribute;
        props.push({ name: name, value: prop.value });
      }
    }

    if (elem.className) {
        props.push({ name: "class", value: elem.className });
    }

    return props.length ? stringify(props) : ""
}

function escapeText(s) {
    var str = '';

    if (typeof(s) === 'string') { 
        str = s; 
    } else if (s) {
        str = s.toString();
    }

    return str
        .replace(/&/g, "&amp;")
        .replace(/</g, "&lt;")
        .replace(/>/g, "&gt;")
}

function escapeAttributeValue(str) {
    return escapeText(str).replace(/"/g, "&quot;")
}

var domWalk$1 = index$3;
var dispatchEvent$1 = dispatchEvent_1;
var addEventListener$1 = addEventListener_1;
var removeEventListener$1 = removeEventListener_1;
var serializeNode = serialize;

var htmlns = "http://www.w3.org/1999/xhtml";

var domElement = DOMElement$1;

function DOMElement$1(tagName, owner, namespace) {
    if (!(this instanceof DOMElement$1)) {
        return new DOMElement$1(tagName)
    }

    var ns = namespace === undefined ? htmlns : (namespace || null);

    this.tagName = ns === htmlns ? String(tagName).toUpperCase() : tagName;
    this.nodeName = this.tagName;
    this.className = "";
    this.dataset = {};
    this.childNodes = [];
    this.parentNode = null;
    this.style = {};
    this.ownerDocument = owner || null;
    this.namespaceURI = ns;
    this._attributes = {};

    if (this.tagName === 'INPUT') {
      this.type = 'text';
    }
}

DOMElement$1.prototype.type = "DOMElement";
DOMElement$1.prototype.nodeType = 1;

DOMElement$1.prototype.appendChild = function _Element_appendChild(child) {
    if (child.parentNode) {
        child.parentNode.removeChild(child);
    }

    this.childNodes.push(child);
    child.parentNode = this;

    return child
};

DOMElement$1.prototype.replaceChild =
    function _Element_replaceChild(elem, needle) {
        // TODO: Throw NotFoundError if needle.parentNode !== this

        if (elem.parentNode) {
            elem.parentNode.removeChild(elem);
        }

        var index = this.childNodes.indexOf(needle);

        needle.parentNode = null;
        this.childNodes[index] = elem;
        elem.parentNode = this;

        return needle
    };

DOMElement$1.prototype.removeChild = function _Element_removeChild(elem) {
    // TODO: Throw NotFoundError if elem.parentNode !== this

    var index = this.childNodes.indexOf(elem);
    this.childNodes.splice(index, 1);

    elem.parentNode = null;
    return elem
};

DOMElement$1.prototype.insertBefore =
    function _Element_insertBefore(elem, needle) {
        // TODO: Throw NotFoundError if referenceElement is a dom node
        // and parentNode !== this

        if (elem.parentNode) {
            elem.parentNode.removeChild(elem);
        }

        var index = needle === null || needle === undefined ?
            -1 :
            this.childNodes.indexOf(needle);

        if (index > -1) {
            this.childNodes.splice(index, 0, elem);
        } else {
            this.childNodes.push(elem);
        }

        elem.parentNode = this;
        return elem
    };

DOMElement$1.prototype.setAttributeNS =
    function _Element_setAttributeNS(namespace, name, value) {
        var prefix = null;
        var localName = name;
        var colonPosition = name.indexOf(":");
        if (colonPosition > -1) {
            prefix = name.substr(0, colonPosition);
            localName = name.substr(colonPosition + 1);
        }
        if (this.tagName === 'INPUT' && name === 'type') {
          this.type = value;
        }
        else {
          var attributes = this._attributes[namespace] || (this._attributes[namespace] = {});
          attributes[localName] = {value: value, prefix: prefix};
        }
    };

DOMElement$1.prototype.getAttributeNS =
    function _Element_getAttributeNS(namespace, name) {
        var attributes = this._attributes[namespace];
        var value = attributes && attributes[name] && attributes[name].value;
        if (this.tagName === 'INPUT' && name === 'type') {
          return this.type;
        }
        if (typeof value !== "string") {
            return null
        }
        return value
    };

DOMElement$1.prototype.removeAttributeNS =
    function _Element_removeAttributeNS(namespace, name) {
        var attributes = this._attributes[namespace];
        if (attributes) {
            delete attributes[name];
        }
    };

DOMElement$1.prototype.hasAttributeNS =
    function _Element_hasAttributeNS(namespace, name) {
        var attributes = this._attributes[namespace];
        return !!attributes && name in attributes;
    };

DOMElement$1.prototype.setAttribute = function _Element_setAttribute(name, value) {
    return this.setAttributeNS(null, name, value)
};

DOMElement$1.prototype.getAttribute = function _Element_getAttribute(name) {
    return this.getAttributeNS(null, name)
};

DOMElement$1.prototype.removeAttribute = function _Element_removeAttribute(name) {
    return this.removeAttributeNS(null, name)
};

DOMElement$1.prototype.hasAttribute = function _Element_hasAttribute(name) {
    return this.hasAttributeNS(null, name)
};

DOMElement$1.prototype.removeEventListener = removeEventListener$1;
DOMElement$1.prototype.addEventListener = addEventListener$1;
DOMElement$1.prototype.dispatchEvent = dispatchEvent$1;

// Un-implemented
DOMElement$1.prototype.focus = function _Element_focus() {
    return void 0
};

DOMElement$1.prototype.toString = function _Element_toString() {
    return serializeNode(this)
};

DOMElement$1.prototype.getElementsByClassName = function _Element_getElementsByClassName(classNames) {
    var classes = classNames.split(" ");
    var elems = [];

    domWalk$1(this, function (node) {
        if (node.nodeType === 1) {
            var nodeClassName = node.className || "";
            var nodeClasses = nodeClassName.split(" ");

            if (classes.every(function (item) {
                return nodeClasses.indexOf(item) !== -1
            })) {
                elems.push(node);
            }
        }
    });

    return elems
};

DOMElement$1.prototype.getElementsByTagName = function _Element_getElementsByTagName(tagName) {
    tagName = tagName.toLowerCase();
    var elems = [];

    domWalk$1(this.childNodes, function (node) {
        if (node.nodeType === 1 && (tagName === '*' || node.tagName.toLowerCase() === tagName)) {
            elems.push(node);
        }
    });

    return elems
};

DOMElement$1.prototype.contains = function _Element_contains(element) {
    return domWalk$1(this, function (node) {
        return element === node
    }) || false
};

var DOMElement$2 = domElement;

var domFragment = DocumentFragment$1;

function DocumentFragment$1(owner) {
    if (!(this instanceof DocumentFragment$1)) {
        return new DocumentFragment$1()
    }

    this.childNodes = [];
    this.parentNode = null;
    this.ownerDocument = owner || null;
}

DocumentFragment$1.prototype.type = "DocumentFragment";
DocumentFragment$1.prototype.nodeType = 11;
DocumentFragment$1.prototype.nodeName = "#document-fragment";

DocumentFragment$1.prototype.appendChild  = DOMElement$2.prototype.appendChild;
DocumentFragment$1.prototype.replaceChild = DOMElement$2.prototype.replaceChild;
DocumentFragment$1.prototype.removeChild  = DOMElement$2.prototype.removeChild;

DocumentFragment$1.prototype.toString =
    function _DocumentFragment_toString() {
        return this.childNodes.map(function (node) {
            return String(node)
        }).join("")
    };

var event = Event$1;

function Event$1(family) {}

Event$1.prototype.initEvent = function _Event_initEvent(type, bubbles, cancelable) {
    this.type = type;
    this.bubbles = bubbles;
    this.cancelable = cancelable;
};

Event$1.prototype.preventDefault = function _Event_preventDefault() {
    
};

var domWalk = index$3;

var Comment = domComment;
var DOMText = domText;
var DOMElement = domElement;
var DocumentFragment = domFragment;
var Event = event;
var dispatchEvent = dispatchEvent_1;
var addEventListener = addEventListener_1;
var removeEventListener = removeEventListener_1;

var document$1 = Document$1;

function Document$1() {
    if (!(this instanceof Document$1)) {
        return new Document$1();
    }

    this.head = this.createElement("head");
    this.body = this.createElement("body");
    this.documentElement = this.createElement("html");
    this.documentElement.appendChild(this.head);
    this.documentElement.appendChild(this.body);
    this.childNodes = [this.documentElement];
    this.nodeType = 9;
}

var proto = Document$1.prototype;
proto.createTextNode = function createTextNode(value) {
    return new DOMText(value, this)
};

proto.createElementNS = function createElementNS(namespace, tagName) {
    var ns = namespace === null ? null : String(namespace);
    return new DOMElement(tagName, this, ns)
};

proto.createElement = function createElement(tagName) {
    return new DOMElement(tagName, this)
};

proto.createDocumentFragment = function createDocumentFragment() {
    return new DocumentFragment(this)
};

proto.createEvent = function createEvent(family) {
    return new Event(family)
};

proto.createComment = function createComment(data) {
    return new Comment(data, this)
};

proto.getElementById = function getElementById(id) {
    id = String(id);

    var result = domWalk(this.childNodes, function (node) {
        if (String(node.id) === id) {
            return node
        }
    });

    return result || null
};

proto.getElementsByClassName = DOMElement.prototype.getElementsByClassName;
proto.getElementsByTagName = DOMElement.prototype.getElementsByTagName;
proto.contains = DOMElement.prototype.contains;

proto.removeEventListener = removeEventListener;
proto.addEventListener = addEventListener;
proto.dispatchEvent = dispatchEvent;

var Document = document$1;

var index$1 = new Document();

var NativeModules = {};

/**
 * 未实现
 * 原本构想是将一些模块拆分出来，注册了才能调用，然而后来需求并没有需要
 * 
 */
function registNativeModule() {
	var mockNativeModules = {
		Http: HttpModule
	};

	for (var module in mockNativeModules) {
		if (isMesh) {
			NativeModules[module] = mesh[module];
		} else {
			NativeModules[module] = mockNativeModules[module];
		}
	}
}

/**
 * 供客户端调用，用来在组件上触发事件，根据 moduleId 索引
 * @param {} moduleId 
 * @param {} event 
 * @param {} params 
 */
function GestureEventHandler(moduleId, event, params) {
	var module = MODULE_MAP[moduleId];
	if (module && module.data && module.data.on) {
		var handler = module.data.on[event];
		if (isFunction(handler)) {
			handler.apply(undefined, toConsumableArray(params));
		}
	}
}

/**
 *  供客户端调用，根据 moduleId 索引找到对应的 Component 实例触发 componentWillShow
 * @param {} moduleId 
 */
function pageWillShow(moduleId) {
	var module = MODULE_MAP[moduleId];
	if (module && module.instance) {
		var instance = module.instance._root;
		instance.componentWillShow && instance.componentWillShow();
		instance._parent && instance._parent.componentWillShow && instance._parent.componentWillShow();
	}
}

/**
 *  供客户端调用，根据 moduleId 索引找到对应的 Component 实例触发 componentWillUnmount
 * @param {} moduleId 
 */
function destroyComponent(moduleId) {
	var module = MODULE_MAP[moduleId];
	if (module && module.instance) {
		var instance = module.instance;
		instance.componentWillUnmount && instance.componentWillUnmount();
		instance._parent && instance._parent.componentWillUnmount && instance._parent.componentWillUnmount();
		delete MODULE_MAP[moduleId];
		instance._unmounted = true;
	}
}

/**
 *  未实现
 * @param {} task 
 */
function TaskModule(task) {
	var module = NativeModules[task.module];
	var method = module[task.method];
	if (!isFunction(task.args[task.args.length - 1])) {
		return new Promise(function (resolve, reject) {
			task.args.push(function (err, data) {
				if (!err) {
					return resolve(data);
				}
				return reject(err);
			});
			method.apply(null, task.args);
		});
	} else {
		method.apply(null, task.args);
	}
}

/**
 * require 方法， 未实现
 * @param {*} moduleName 
 */
function requireModule(moduleName) {
	var methods = NativeModules[moduleName];
	var target = {};
	if (!methods) {
		throwError(moduleName + ' module not exist');
		return;
	}

	var _loop = function _loop(methodName) {
		Object.defineProperty(target, methodName, {
			configurable: false,
			enumerable: true,
			get: function moduleGetter() {
				return function () {
					for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
						args[_key] = arguments[_key];
					}

					return TaskModule({
						module: moduleName,
						method: methodName,
						args: args
					});
				};
			}
		});
	};

	for (var methodName in methods) {
		_loop(methodName);
	}
	return target;
}

/**
 * http Module, 未实现
 */
var HttpModule = {
	get: function get$$1(url, callback) {
		callback(null, url);
	}
};

var es6Promise = createCommonjsModule(function (module, exports) {
/*!
 * @overview es6-promise - a tiny implementation of Promises/A+.
 * @copyright Copyright (c) 2014 Yehuda Katz, Tom Dale, Stefan Penner and contributors (Conversion to ES6 API by Jake Archibald)
 * @license   Licensed under MIT license
 *            See https://raw.githubusercontent.com/stefanpenner/es6-promise/master/LICENSE
 * @version   4.1.0
 */

(function (global, factory) {
    module.exports = factory();
}(commonjsGlobal, (function () { 'use strict';

function objectOrFunction(x) {
  return typeof x === 'function' || typeof x === 'object' && x !== null;
}

function isFunction$$1(x) {
  return typeof x === 'function';
}

var _isArray = undefined;
if (!Array.isArray) {
  _isArray = function (x) {
    return Object.prototype.toString.call(x) === '[object Array]';
  };
} else {
  _isArray = Array.isArray;
}

var isArray$$1 = _isArray;

var len = 0;
var vertxNext = undefined;
var customSchedulerFn = undefined;

var asap = function asap(callback, arg) {
  queue[len] = callback;
  queue[len + 1] = arg;
  len += 2;
  if (len === 2) {
    // If len is 2, that means that we need to schedule an async flush.
    // If additional callbacks are queued before the queue is flushed, they
    // will be processed by this flush that we are scheduling.
    if (customSchedulerFn) {
      customSchedulerFn(flush);
    } else {
      scheduleFlush();
    }
  }
};

function setScheduler(scheduleFn) {
  customSchedulerFn = scheduleFn;
}

function setAsap(asapFn) {
  asap = asapFn;
}

var browserWindow = typeof window !== 'undefined' ? window : undefined;
var browserGlobal = browserWindow || {};
var BrowserMutationObserver = browserGlobal.MutationObserver || browserGlobal.WebKitMutationObserver;
var isNode = typeof self === 'undefined' && typeof process !== 'undefined' && ({}).toString.call(process) === '[object process]';

// test for web worker but not in IE10
var isWorker = typeof Uint8ClampedArray !== 'undefined' && typeof importScripts !== 'undefined' && typeof MessageChannel !== 'undefined';

// node
function useNextTick() {
  // node version 0.10.x displays a deprecation warning when nextTick is used recursively
  // see https://github.com/cujojs/when/issues/410 for details
  return function () {
    return process.nextTick(flush);
  };
}

// vertx
function useVertxTimer() {
  if (typeof vertxNext !== 'undefined') {
    return function () {
      vertxNext(flush);
    };
  }

  return useSetTimeout();
}

function useMutationObserver() {
  var iterations = 0;
  var observer = new BrowserMutationObserver(flush);
  var node = document.createTextNode('');
  observer.observe(node, { characterData: true });

  return function () {
    node.data = iterations = ++iterations % 2;
  };
}

// web worker
function useMessageChannel() {
  var channel = new MessageChannel();
  channel.port1.onmessage = flush;
  return function () {
    return channel.port2.postMessage(0);
  };
}

function useSetTimeout() {
  // Store setTimeout reference so es6-promise will be unaffected by
  // other code modifying setTimeout (like sinon.useFakeTimers())
  var globalSetTimeout = setTimeout;
  return function () {
    return globalSetTimeout(flush, 1);
  };
}

var queue = new Array(1000);
function flush() {
  for (var i = 0; i < len; i += 2) {
    var callback = queue[i];
    var arg = queue[i + 1];

    callback(arg);

    queue[i] = undefined;
    queue[i + 1] = undefined;
  }

  len = 0;
}

function attemptVertx() {
  try {
    var r = commonjsRequire;
    var vertx = r('vertx');
    vertxNext = vertx.runOnLoop || vertx.runOnContext;
    return useVertxTimer();
  } catch (e) {
    return useSetTimeout();
  }
}

var scheduleFlush = undefined;
// Decide what async method to use to triggering processing of queued callbacks:
if (isNode) {
  scheduleFlush = useNextTick();
} else if (BrowserMutationObserver) {
  scheduleFlush = useMutationObserver();
} else if (isWorker) {
  scheduleFlush = useMessageChannel();
} else if (browserWindow === undefined && typeof commonjsRequire === 'function') {
  scheduleFlush = attemptVertx();
} else {
  scheduleFlush = useSetTimeout();
}

function then(onFulfillment, onRejection) {
  var _arguments = arguments;

  var parent = this;

  var child = new this.constructor(noop);

  if (child[PROMISE_ID] === undefined) {
    makePromise(child);
  }

  var _state = parent._state;

  if (_state) {
    (function () {
      var callback = _arguments[_state - 1];
      asap(function () {
        return invokeCallback(_state, child, callback, parent._result);
      });
    })();
  } else {
    subscribe(parent, child, onFulfillment, onRejection);
  }

  return child;
}

/**
  `Promise.resolve` returns a promise that will become resolved with the
  passed `value`. It is shorthand for the following:

  ```javascript
  let promise = new Promise(function(resolve, reject){
    resolve(1);
  });

  promise.then(function(value){
    // value === 1
  });
  ```

  Instead of writing the above, your code now simply becomes the following:

  ```javascript
  let promise = Promise.resolve(1);

  promise.then(function(value){
    // value === 1
  });
  ```

  @method resolve
  @static
  @param {Any} value value that the returned promise will be resolved with
  Useful for tooling.
  @return {Promise} a promise that will become fulfilled with the given
  `value`
*/
function resolve(object) {
  /*jshint validthis:true */
  var Constructor = this;

  if (object && typeof object === 'object' && object.constructor === Constructor) {
    return object;
  }

  var promise = new Constructor(noop);
  _resolve(promise, object);
  return promise;
}

var PROMISE_ID = Math.random().toString(36).substring(16);

function noop() {}

var PENDING = void 0;
var FULFILLED = 1;
var REJECTED = 2;

var GET_THEN_ERROR = new ErrorObject();

function selfFulfillment() {
  return new TypeError("You cannot resolve a promise with itself");
}

function cannotReturnOwn() {
  return new TypeError('A promises callback cannot return that same promise.');
}

function getThen(promise) {
  try {
    return promise.then;
  } catch (error) {
    GET_THEN_ERROR.error = error;
    return GET_THEN_ERROR;
  }
}

function tryThen(then, value, fulfillmentHandler, rejectionHandler) {
  try {
    then.call(value, fulfillmentHandler, rejectionHandler);
  } catch (e) {
    return e;
  }
}

function handleForeignThenable(promise, thenable, then) {
  asap(function (promise) {
    var sealed = false;
    var error = tryThen(then, thenable, function (value) {
      if (sealed) {
        return;
      }
      sealed = true;
      if (thenable !== value) {
        _resolve(promise, value);
      } else {
        fulfill(promise, value);
      }
    }, function (reason) {
      if (sealed) {
        return;
      }
      sealed = true;

      _reject(promise, reason);
    }, 'Settle: ' + (promise._label || ' unknown promise'));

    if (!sealed && error) {
      sealed = true;
      _reject(promise, error);
    }
  }, promise);
}

function handleOwnThenable(promise, thenable) {
  if (thenable._state === FULFILLED) {
    fulfill(promise, thenable._result);
  } else if (thenable._state === REJECTED) {
    _reject(promise, thenable._result);
  } else {
    subscribe(thenable, undefined, function (value) {
      return _resolve(promise, value);
    }, function (reason) {
      return _reject(promise, reason);
    });
  }
}

function handleMaybeThenable(promise, maybeThenable, then$$) {
  if (maybeThenable.constructor === promise.constructor && then$$ === then && maybeThenable.constructor.resolve === resolve) {
    handleOwnThenable(promise, maybeThenable);
  } else {
    if (then$$ === GET_THEN_ERROR) {
      _reject(promise, GET_THEN_ERROR.error);
      GET_THEN_ERROR.error = null;
    } else if (then$$ === undefined) {
      fulfill(promise, maybeThenable);
    } else if (isFunction$$1(then$$)) {
      handleForeignThenable(promise, maybeThenable, then$$);
    } else {
      fulfill(promise, maybeThenable);
    }
  }
}

function _resolve(promise, value) {
  if (promise === value) {
    _reject(promise, selfFulfillment());
  } else if (objectOrFunction(value)) {
    handleMaybeThenable(promise, value, getThen(value));
  } else {
    fulfill(promise, value);
  }
}

function publishRejection(promise) {
  if (promise._onerror) {
    promise._onerror(promise._result);
  }

  publish(promise);
}

function fulfill(promise, value) {
  if (promise._state !== PENDING) {
    return;
  }

  promise._result = value;
  promise._state = FULFILLED;

  if (promise._subscribers.length !== 0) {
    asap(publish, promise);
  }
}

function _reject(promise, reason) {
  if (promise._state !== PENDING) {
    return;
  }
  promise._state = REJECTED;
  promise._result = reason;

  asap(publishRejection, promise);
}

function subscribe(parent, child, onFulfillment, onRejection) {
  var _subscribers = parent._subscribers;
  var length = _subscribers.length;

  parent._onerror = null;

  _subscribers[length] = child;
  _subscribers[length + FULFILLED] = onFulfillment;
  _subscribers[length + REJECTED] = onRejection;

  if (length === 0 && parent._state) {
    asap(publish, parent);
  }
}

function publish(promise) {
  var subscribers = promise._subscribers;
  var settled = promise._state;

  if (subscribers.length === 0) {
    return;
  }

  var child = undefined,
      callback = undefined,
      detail = promise._result;

  for (var i = 0; i < subscribers.length; i += 3) {
    child = subscribers[i];
    callback = subscribers[i + settled];

    if (child) {
      invokeCallback(settled, child, callback, detail);
    } else {
      callback(detail);
    }
  }

  promise._subscribers.length = 0;
}

function ErrorObject() {
  this.error = null;
}

var TRY_CATCH_ERROR = new ErrorObject();

function tryCatch(callback, detail) {
  try {
    return callback(detail);
  } catch (e) {
    TRY_CATCH_ERROR.error = e;
    return TRY_CATCH_ERROR;
  }
}

function invokeCallback(settled, promise, callback, detail) {
  var hasCallback = isFunction$$1(callback),
      value = undefined,
      error = undefined,
      succeeded = undefined,
      failed = undefined;

  if (hasCallback) {
    value = tryCatch(callback, detail);

    if (value === TRY_CATCH_ERROR) {
      failed = true;
      error = value.error;
      value.error = null;
    } else {
      succeeded = true;
    }

    if (promise === value) {
      _reject(promise, cannotReturnOwn());
      return;
    }
  } else {
    value = detail;
    succeeded = true;
  }

  if (promise._state !== PENDING) {
    // noop
  } else if (hasCallback && succeeded) {
      _resolve(promise, value);
    } else if (failed) {
      _reject(promise, error);
    } else if (settled === FULFILLED) {
      fulfill(promise, value);
    } else if (settled === REJECTED) {
      _reject(promise, value);
    }
}

function initializePromise(promise, resolver) {
  try {
    resolver(function resolvePromise(value) {
      _resolve(promise, value);
    }, function rejectPromise(reason) {
      _reject(promise, reason);
    });
  } catch (e) {
    _reject(promise, e);
  }
}

var id = 0;
function nextId() {
  return id++;
}

function makePromise(promise) {
  promise[PROMISE_ID] = id++;
  promise._state = undefined;
  promise._result = undefined;
  promise._subscribers = [];
}

function Enumerator(Constructor, input) {
  this._instanceConstructor = Constructor;
  this.promise = new Constructor(noop);

  if (!this.promise[PROMISE_ID]) {
    makePromise(this.promise);
  }

  if (isArray$$1(input)) {
    this._input = input;
    this.length = input.length;
    this._remaining = input.length;

    this._result = new Array(this.length);

    if (this.length === 0) {
      fulfill(this.promise, this._result);
    } else {
      this.length = this.length || 0;
      this._enumerate();
      if (this._remaining === 0) {
        fulfill(this.promise, this._result);
      }
    }
  } else {
    _reject(this.promise, validationError());
  }
}

function validationError() {
  return new Error('Array Methods must be provided an Array');
}

Enumerator.prototype._enumerate = function () {
  var length = this.length;
  var _input = this._input;

  for (var i = 0; this._state === PENDING && i < length; i++) {
    this._eachEntry(_input[i], i);
  }
};

Enumerator.prototype._eachEntry = function (entry, i) {
  var c = this._instanceConstructor;
  var resolve$$ = c.resolve;

  if (resolve$$ === resolve) {
    var _then = getThen(entry);

    if (_then === then && entry._state !== PENDING) {
      this._settledAt(entry._state, i, entry._result);
    } else if (typeof _then !== 'function') {
      this._remaining--;
      this._result[i] = entry;
    } else if (c === Promise) {
      var promise = new c(noop);
      handleMaybeThenable(promise, entry, _then);
      this._willSettleAt(promise, i);
    } else {
      this._willSettleAt(new c(function (resolve$$) {
        return resolve$$(entry);
      }), i);
    }
  } else {
    this._willSettleAt(resolve$$(entry), i);
  }
};

Enumerator.prototype._settledAt = function (state, i, value) {
  var promise = this.promise;

  if (promise._state === PENDING) {
    this._remaining--;

    if (state === REJECTED) {
      _reject(promise, value);
    } else {
      this._result[i] = value;
    }
  }

  if (this._remaining === 0) {
    fulfill(promise, this._result);
  }
};

Enumerator.prototype._willSettleAt = function (promise, i) {
  var enumerator = this;

  subscribe(promise, undefined, function (value) {
    return enumerator._settledAt(FULFILLED, i, value);
  }, function (reason) {
    return enumerator._settledAt(REJECTED, i, reason);
  });
};

/**
  `Promise.all` accepts an array of promises, and returns a new promise which
  is fulfilled with an array of fulfillment values for the passed promises, or
  rejected with the reason of the first passed promise to be rejected. It casts all
  elements of the passed iterable to promises as it runs this algorithm.

  Example:

  ```javascript
  let promise1 = resolve(1);
  let promise2 = resolve(2);
  let promise3 = resolve(3);
  let promises = [ promise1, promise2, promise3 ];

  Promise.all(promises).then(function(array){
    // The array here would be [ 1, 2, 3 ];
  });
  ```

  If any of the `promises` given to `all` are rejected, the first promise
  that is rejected will be given as an argument to the returned promises's
  rejection handler. For example:

  Example:

  ```javascript
  let promise1 = resolve(1);
  let promise2 = reject(new Error("2"));
  let promise3 = reject(new Error("3"));
  let promises = [ promise1, promise2, promise3 ];

  Promise.all(promises).then(function(array){
    // Code here never runs because there are rejected promises!
  }, function(error) {
    // error.message === "2"
  });
  ```

  @method all
  @static
  @param {Array} entries array of promises
  @param {String} label optional string for labeling the promise.
  Useful for tooling.
  @return {Promise} promise that is fulfilled when all `promises` have been
  fulfilled, or rejected if any of them become rejected.
  @static
*/
function all(entries) {
  return new Enumerator(this, entries).promise;
}

/**
  `Promise.race` returns a new promise which is settled in the same way as the
  first passed promise to settle.

  Example:

  ```javascript
  let promise1 = new Promise(function(resolve, reject){
    setTimeout(function(){
      resolve('promise 1');
    }, 200);
  });

  let promise2 = new Promise(function(resolve, reject){
    setTimeout(function(){
      resolve('promise 2');
    }, 100);
  });

  Promise.race([promise1, promise2]).then(function(result){
    // result === 'promise 2' because it was resolved before promise1
    // was resolved.
  });
  ```

  `Promise.race` is deterministic in that only the state of the first
  settled promise matters. For example, even if other promises given to the
  `promises` array argument are resolved, but the first settled promise has
  become rejected before the other promises became fulfilled, the returned
  promise will become rejected:

  ```javascript
  let promise1 = new Promise(function(resolve, reject){
    setTimeout(function(){
      resolve('promise 1');
    }, 200);
  });

  let promise2 = new Promise(function(resolve, reject){
    setTimeout(function(){
      reject(new Error('promise 2'));
    }, 100);
  });

  Promise.race([promise1, promise2]).then(function(result){
    // Code here never runs
  }, function(reason){
    // reason.message === 'promise 2' because promise 2 became rejected before
    // promise 1 became fulfilled
  });
  ```

  An example real-world use case is implementing timeouts:

  ```javascript
  Promise.race([ajax('foo.json'), timeout(5000)])
  ```

  @method race
  @static
  @param {Array} promises array of promises to observe
  Useful for tooling.
  @return {Promise} a promise which settles in the same way as the first passed
  promise to settle.
*/
function race(entries) {
  /*jshint validthis:true */
  var Constructor = this;

  if (!isArray$$1(entries)) {
    return new Constructor(function (_, reject) {
      return reject(new TypeError('You must pass an array to race.'));
    });
  } else {
    return new Constructor(function (resolve, reject) {
      var length = entries.length;
      for (var i = 0; i < length; i++) {
        Constructor.resolve(entries[i]).then(resolve, reject);
      }
    });
  }
}

/**
  `Promise.reject` returns a promise rejected with the passed `reason`.
  It is shorthand for the following:

  ```javascript
  let promise = new Promise(function(resolve, reject){
    reject(new Error('WHOOPS'));
  });

  promise.then(function(value){
    // Code here doesn't run because the promise is rejected!
  }, function(reason){
    // reason.message === 'WHOOPS'
  });
  ```

  Instead of writing the above, your code now simply becomes the following:

  ```javascript
  let promise = Promise.reject(new Error('WHOOPS'));

  promise.then(function(value){
    // Code here doesn't run because the promise is rejected!
  }, function(reason){
    // reason.message === 'WHOOPS'
  });
  ```

  @method reject
  @static
  @param {Any} reason value that the returned promise will be rejected with.
  Useful for tooling.
  @return {Promise} a promise rejected with the given `reason`.
*/
function reject(reason) {
  /*jshint validthis:true */
  var Constructor = this;
  var promise = new Constructor(noop);
  _reject(promise, reason);
  return promise;
}

function needsResolver() {
  throw new TypeError('You must pass a resolver function as the first argument to the promise constructor');
}

function needsNew() {
  throw new TypeError("Failed to construct 'Promise': Please use the 'new' operator, this object constructor cannot be called as a function.");
}

/**
  Promise objects represent the eventual result of an asynchronous operation. The
  primary way of interacting with a promise is through its `then` method, which
  registers callbacks to receive either a promise's eventual value or the reason
  why the promise cannot be fulfilled.

  Terminology
  -----------

  - `promise` is an object or function with a `then` method whose behavior conforms to this specification.
  - `thenable` is an object or function that defines a `then` method.
  - `value` is any legal JavaScript value (including undefined, a thenable, or a promise).
  - `exception` is a value that is thrown using the throw statement.
  - `reason` is a value that indicates why a promise was rejected.
  - `settled` the final resting state of a promise, fulfilled or rejected.

  A promise can be in one of three states: pending, fulfilled, or rejected.

  Promises that are fulfilled have a fulfillment value and are in the fulfilled
  state.  Promises that are rejected have a rejection reason and are in the
  rejected state.  A fulfillment value is never a thenable.

  Promises can also be said to *resolve* a value.  If this value is also a
  promise, then the original promise's settled state will match the value's
  settled state.  So a promise that *resolves* a promise that rejects will
  itself reject, and a promise that *resolves* a promise that fulfills will
  itself fulfill.


  Basic Usage:
  ------------

  ```js
  let promise = new Promise(function(resolve, reject) {
    // on success
    resolve(value);

    // on failure
    reject(reason);
  });

  promise.then(function(value) {
    // on fulfillment
  }, function(reason) {
    // on rejection
  });
  ```

  Advanced Usage:
  ---------------

  Promises shine when abstracting away asynchronous interactions such as
  `XMLHttpRequest`s.

  ```js
  function getJSON(url) {
    return new Promise(function(resolve, reject){
      let xhr = new XMLHttpRequest();

      xhr.open('GET', url);
      xhr.onreadystatechange = handler;
      xhr.responseType = 'json';
      xhr.setRequestHeader('Accept', 'application/json');
      xhr.send();

      function handler() {
        if (this.readyState === this.DONE) {
          if (this.status === 200) {
            resolve(this.response);
          } else {
            reject(new Error('getJSON: `' + url + '` failed with status: [' + this.status + ']'));
          }
        }
      };
    });
  }

  getJSON('/posts.json').then(function(json) {
    // on fulfillment
  }, function(reason) {
    // on rejection
  });
  ```

  Unlike callbacks, promises are great composable primitives.

  ```js
  Promise.all([
    getJSON('/posts'),
    getJSON('/comments')
  ]).then(function(values){
    values[0] // => postsJSON
    values[1] // => commentsJSON

    return values;
  });
  ```

  @class Promise
  @param {function} resolver
  Useful for tooling.
  @constructor
*/
function Promise(resolver) {
  this[PROMISE_ID] = nextId();
  this._result = this._state = undefined;
  this._subscribers = [];

  if (noop !== resolver) {
    typeof resolver !== 'function' && needsResolver();
    this instanceof Promise ? initializePromise(this, resolver) : needsNew();
  }
}

Promise.all = all;
Promise.race = race;
Promise.resolve = resolve;
Promise.reject = reject;
Promise._setScheduler = setScheduler;
Promise._setAsap = setAsap;
Promise._asap = asap;

Promise.prototype = {
  constructor: Promise,

  /**
    The primary way of interacting with a promise is through its `then` method,
    which registers callbacks to receive either a promise's eventual value or the
    reason why the promise cannot be fulfilled.
  
    ```js
    findUser().then(function(user){
      // user is available
    }, function(reason){
      // user is unavailable, and you are given the reason why
    });
    ```
  
    Chaining
    --------
  
    The return value of `then` is itself a promise.  This second, 'downstream'
    promise is resolved with the return value of the first promise's fulfillment
    or rejection handler, or rejected if the handler throws an exception.
  
    ```js
    findUser().then(function (user) {
      return user.name;
    }, function (reason) {
      return 'default name';
    }).then(function (userName) {
      // If `findUser` fulfilled, `userName` will be the user's name, otherwise it
      // will be `'default name'`
    });
  
    findUser().then(function (user) {
      throw new Error('Found user, but still unhappy');
    }, function (reason) {
      throw new Error('`findUser` rejected and we're unhappy');
    }).then(function (value) {
      // never reached
    }, function (reason) {
      // if `findUser` fulfilled, `reason` will be 'Found user, but still unhappy'.
      // If `findUser` rejected, `reason` will be '`findUser` rejected and we're unhappy'.
    });
    ```
    If the downstream promise does not specify a rejection handler, rejection reasons will be propagated further downstream.
  
    ```js
    findUser().then(function (user) {
      throw new PedagogicalException('Upstream error');
    }).then(function (value) {
      // never reached
    }).then(function (value) {
      // never reached
    }, function (reason) {
      // The `PedgagocialException` is propagated all the way down to here
    });
    ```
  
    Assimilation
    ------------
  
    Sometimes the value you want to propagate to a downstream promise can only be
    retrieved asynchronously. This can be achieved by returning a promise in the
    fulfillment or rejection handler. The downstream promise will then be pending
    until the returned promise is settled. This is called *assimilation*.
  
    ```js
    findUser().then(function (user) {
      return findCommentsByAuthor(user);
    }).then(function (comments) {
      // The user's comments are now available
    });
    ```
  
    If the assimliated promise rejects, then the downstream promise will also reject.
  
    ```js
    findUser().then(function (user) {
      return findCommentsByAuthor(user);
    }).then(function (comments) {
      // If `findCommentsByAuthor` fulfills, we'll have the value here
    }, function (reason) {
      // If `findCommentsByAuthor` rejects, we'll have the reason here
    });
    ```
  
    Simple Example
    --------------
  
    Synchronous Example
  
    ```javascript
    let result;
  
    try {
      result = findResult();
      // success
    } catch(reason) {
      // failure
    }
    ```
  
    Errback Example
  
    ```js
    findResult(function(result, err){
      if (err) {
        // failure
      } else {
        // success
      }
    });
    ```
  
    Promise Example;
  
    ```javascript
    findResult().then(function(result){
      // success
    }, function(reason){
      // failure
    });
    ```
  
    Advanced Example
    --------------
  
    Synchronous Example
  
    ```javascript
    let author, books;
  
    try {
      author = findAuthor();
      books  = findBooksByAuthor(author);
      // success
    } catch(reason) {
      // failure
    }
    ```
  
    Errback Example
  
    ```js
  
    function foundBooks(books) {
  
    }
  
    function failure(reason) {
  
    }
  
    findAuthor(function(author, err){
      if (err) {
        failure(err);
        // failure
      } else {
        try {
          findBoooksByAuthor(author, function(books, err) {
            if (err) {
              failure(err);
            } else {
              try {
                foundBooks(books);
              } catch(reason) {
                failure(reason);
              }
            }
          });
        } catch(error) {
          failure(err);
        }
        // success
      }
    });
    ```
  
    Promise Example;
  
    ```javascript
    findAuthor().
      then(findBooksByAuthor).
      then(function(books){
        // found books
    }).catch(function(reason){
      // something went wrong
    });
    ```
  
    @method then
    @param {Function} onFulfilled
    @param {Function} onRejected
    Useful for tooling.
    @return {Promise}
  */
  then: then,

  /**
    `catch` is simply sugar for `then(undefined, onRejection)` which makes it the same
    as the catch block of a try/catch statement.
  
    ```js
    function findAuthor(){
      throw new Error('couldn't find that author');
    }
  
    // synchronous
    try {
      findAuthor();
    } catch(reason) {
      // something went wrong
    }
  
    // async with promises
    findAuthor().catch(function(reason){
      // something went wrong
    });
    ```
  
    @method catch
    @param {Function} onRejection
    Useful for tooling.
    @return {Promise}
  */
  'catch': function _catch(onRejection) {
    return this.then(null, onRejection);
  }
};

function polyfill() {
    var local = undefined;

    if (typeof commonjsGlobal !== 'undefined') {
        local = commonjsGlobal;
    } else if (typeof self !== 'undefined') {
        local = self;
    } else {
        try {
            local = Function('return this')();
        } catch (e) {
            throw new Error('polyfill failed because global object is unavailable in this environment');
        }
    }

    var P = local.Promise;

    if (P) {
        var promiseToString = null;
        try {
            promiseToString = Object.prototype.toString.call(P.resolve());
        } catch (e) {
            // silently ignored
        }

        if (promiseToString === '[object Promise]' && !P.cast) {
            return;
        }
    }

    local.Promise = Promise;
}

// Strange compat..
Promise.polyfill = polyfill;
Promise.Promise = Promise;

return Promise;

})));

});

var runtime = createCommonjsModule(function (module) {
/**
 * Copyright (c) 2014-present, Facebook, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

!(function(global) {
  "use strict";

  var Op = Object.prototype;
  var hasOwn = Op.hasOwnProperty;
  var undefined; // More compressible than void 0.
  var $Symbol = typeof Symbol === "function" ? Symbol : {};
  var iteratorSymbol = $Symbol.iterator || "@@iterator";
  var asyncIteratorSymbol = $Symbol.asyncIterator || "@@asyncIterator";
  var toStringTagSymbol = $Symbol.toStringTag || "@@toStringTag";

  var inModule = 'object' === "object";
  var runtime = global.regeneratorRuntime;
  if (runtime) {
    if (inModule) {
      // If regeneratorRuntime is defined globally and we're in a module,
      // make the exports object identical to regeneratorRuntime.
      module.exports = runtime;
    }
    // Don't bother evaluating the rest of this file if the runtime was
    // already defined globally.
    return;
  }

  // Define the runtime globally (as expected by generated code) as either
  // module.exports (if we're in a module) or a new, empty object.
  runtime = global.regeneratorRuntime = inModule ? module.exports : {};

  function wrap(innerFn, outerFn, self, tryLocsList) {
    // If outerFn provided and outerFn.prototype is a Generator, then outerFn.prototype instanceof Generator.
    var protoGenerator = outerFn && outerFn.prototype instanceof Generator ? outerFn : Generator;
    var generator = Object.create(protoGenerator.prototype);
    var context = new Context(tryLocsList || []);

    // The ._invoke method unifies the implementations of the .next,
    // .throw, and .return methods.
    generator._invoke = makeInvokeMethod(innerFn, self, context);

    return generator;
  }
  runtime.wrap = wrap;

  // Try/catch helper to minimize deoptimizations. Returns a completion
  // record like context.tryEntries[i].completion. This interface could
  // have been (and was previously) designed to take a closure to be
  // invoked without arguments, but in all the cases we care about we
  // already have an existing method we want to call, so there's no need
  // to create a new function object. We can even get away with assuming
  // the method takes exactly one argument, since that happens to be true
  // in every case, so we don't have to touch the arguments object. The
  // only additional allocation required is the completion record, which
  // has a stable shape and so hopefully should be cheap to allocate.
  function tryCatch(fn, obj, arg) {
    try {
      return { type: "normal", arg: fn.call(obj, arg) };
    } catch (err) {
      return { type: "throw", arg: err };
    }
  }

  var GenStateSuspendedStart = "suspendedStart";
  var GenStateSuspendedYield = "suspendedYield";
  var GenStateExecuting = "executing";
  var GenStateCompleted = "completed";

  // Returning this object from the innerFn has the same effect as
  // breaking out of the dispatch switch statement.
  var ContinueSentinel = {};

  // Dummy constructor functions that we use as the .constructor and
  // .constructor.prototype properties for functions that return Generator
  // objects. For full spec compliance, you may wish to configure your
  // minifier not to mangle the names of these two functions.
  function Generator() {}
  function GeneratorFunction() {}
  function GeneratorFunctionPrototype() {}

  // This is a polyfill for %IteratorPrototype% for environments that
  // don't natively support it.
  var IteratorPrototype = {};
  IteratorPrototype[iteratorSymbol] = function () {
    return this;
  };

  var getProto = Object.getPrototypeOf;
  var NativeIteratorPrototype = getProto && getProto(getProto(values([])));
  if (NativeIteratorPrototype &&
      NativeIteratorPrototype !== Op &&
      hasOwn.call(NativeIteratorPrototype, iteratorSymbol)) {
    // This environment has a native %IteratorPrototype%; use it instead
    // of the polyfill.
    IteratorPrototype = NativeIteratorPrototype;
  }

  var Gp = GeneratorFunctionPrototype.prototype =
    Generator.prototype = Object.create(IteratorPrototype);
  GeneratorFunction.prototype = Gp.constructor = GeneratorFunctionPrototype;
  GeneratorFunctionPrototype.constructor = GeneratorFunction;
  GeneratorFunctionPrototype[toStringTagSymbol] =
    GeneratorFunction.displayName = "GeneratorFunction";

  // Helper for defining the .next, .throw, and .return methods of the
  // Iterator interface in terms of a single ._invoke method.
  function defineIteratorMethods(prototype) {
    ["next", "throw", "return"].forEach(function(method) {
      prototype[method] = function(arg) {
        return this._invoke(method, arg);
      };
    });
  }

  runtime.isGeneratorFunction = function(genFun) {
    var ctor = typeof genFun === "function" && genFun.constructor;
    return ctor
      ? ctor === GeneratorFunction ||
        // For the native GeneratorFunction constructor, the best we can
        // do is to check its .name property.
        (ctor.displayName || ctor.name) === "GeneratorFunction"
      : false;
  };

  runtime.mark = function(genFun) {
    if (Object.setPrototypeOf) {
      Object.setPrototypeOf(genFun, GeneratorFunctionPrototype);
    } else {
      genFun.__proto__ = GeneratorFunctionPrototype;
      if (!(toStringTagSymbol in genFun)) {
        genFun[toStringTagSymbol] = "GeneratorFunction";
      }
    }
    genFun.prototype = Object.create(Gp);
    return genFun;
  };

  // Within the body of any async function, `await x` is transformed to
  // `yield regeneratorRuntime.awrap(x)`, so that the runtime can test
  // `hasOwn.call(value, "__await")` to determine if the yielded value is
  // meant to be awaited.
  runtime.awrap = function(arg) {
    return { __await: arg };
  };

  function AsyncIterator(generator) {
    function invoke(method, arg, resolve, reject) {
      var record = tryCatch(generator[method], generator, arg);
      if (record.type === "throw") {
        reject(record.arg);
      } else {
        var result = record.arg;
        var value = result.value;
        if (value &&
            typeof value === "object" &&
            hasOwn.call(value, "__await")) {
          return Promise.resolve(value.__await).then(function(value) {
            invoke("next", value, resolve, reject);
          }, function(err) {
            invoke("throw", err, resolve, reject);
          });
        }

        return Promise.resolve(value).then(function(unwrapped) {
          // When a yielded Promise is resolved, its final value becomes
          // the .value of the Promise<{value,done}> result for the
          // current iteration. If the Promise is rejected, however, the
          // result for this iteration will be rejected with the same
          // reason. Note that rejections of yielded Promises are not
          // thrown back into the generator function, as is the case
          // when an awaited Promise is rejected. This difference in
          // behavior between yield and await is important, because it
          // allows the consumer to decide what to do with the yielded
          // rejection (swallow it and continue, manually .throw it back
          // into the generator, abandon iteration, whatever). With
          // await, by contrast, there is no opportunity to examine the
          // rejection reason outside the generator function, so the
          // only option is to throw it from the await expression, and
          // let the generator function handle the exception.
          result.value = unwrapped;
          resolve(result);
        }, reject);
      }
    }

    var previousPromise;

    function enqueue(method, arg) {
      function callInvokeWithMethodAndArg() {
        return new Promise(function(resolve, reject) {
          invoke(method, arg, resolve, reject);
        });
      }

      return previousPromise =
        // If enqueue has been called before, then we want to wait until
        // all previous Promises have been resolved before calling invoke,
        // so that results are always delivered in the correct order. If
        // enqueue has not been called before, then it is important to
        // call invoke immediately, without waiting on a callback to fire,
        // so that the async generator function has the opportunity to do
        // any necessary setup in a predictable way. This predictability
        // is why the Promise constructor synchronously invokes its
        // executor callback, and why async functions synchronously
        // execute code before the first await. Since we implement simple
        // async functions in terms of async generators, it is especially
        // important to get this right, even though it requires care.
        previousPromise ? previousPromise.then(
          callInvokeWithMethodAndArg,
          // Avoid propagating failures to Promises returned by later
          // invocations of the iterator.
          callInvokeWithMethodAndArg
        ) : callInvokeWithMethodAndArg();
    }

    // Define the unified helper method that is used to implement .next,
    // .throw, and .return (see defineIteratorMethods).
    this._invoke = enqueue;
  }

  defineIteratorMethods(AsyncIterator.prototype);
  AsyncIterator.prototype[asyncIteratorSymbol] = function () {
    return this;
  };
  runtime.AsyncIterator = AsyncIterator;

  // Note that simple async functions are implemented on top of
  // AsyncIterator objects; they just return a Promise for the value of
  // the final result produced by the iterator.
  runtime.async = function(innerFn, outerFn, self, tryLocsList) {
    var iter = new AsyncIterator(
      wrap(innerFn, outerFn, self, tryLocsList)
    );

    return runtime.isGeneratorFunction(outerFn)
      ? iter // If outerFn is a generator, return the full iterator.
      : iter.next().then(function(result) {
          return result.done ? result.value : iter.next();
        });
  };

  function makeInvokeMethod(innerFn, self, context) {
    var state = GenStateSuspendedStart;

    return function invoke(method, arg) {
      if (state === GenStateExecuting) {
        throw new Error("Generator is already running");
      }

      if (state === GenStateCompleted) {
        if (method === "throw") {
          throw arg;
        }

        // Be forgiving, per 25.3.3.3.3 of the spec:
        // https://people.mozilla.org/~jorendorff/es6-draft.html#sec-generatorresume
        return doneResult();
      }

      context.method = method;
      context.arg = arg;

      while (true) {
        var delegate = context.delegate;
        if (delegate) {
          var delegateResult = maybeInvokeDelegate(delegate, context);
          if (delegateResult) {
            if (delegateResult === ContinueSentinel) continue;
            return delegateResult;
          }
        }

        if (context.method === "next") {
          // Setting context._sent for legacy support of Babel's
          // function.sent implementation.
          context.sent = context._sent = context.arg;

        } else if (context.method === "throw") {
          if (state === GenStateSuspendedStart) {
            state = GenStateCompleted;
            throw context.arg;
          }

          context.dispatchException(context.arg);

        } else if (context.method === "return") {
          context.abrupt("return", context.arg);
        }

        state = GenStateExecuting;

        var record = tryCatch(innerFn, self, context);
        if (record.type === "normal") {
          // If an exception is thrown from innerFn, we leave state ===
          // GenStateExecuting and loop back for another invocation.
          state = context.done
            ? GenStateCompleted
            : GenStateSuspendedYield;

          if (record.arg === ContinueSentinel) {
            continue;
          }

          return {
            value: record.arg,
            done: context.done
          };

        } else if (record.type === "throw") {
          state = GenStateCompleted;
          // Dispatch the exception by looping back around to the
          // context.dispatchException(context.arg) call above.
          context.method = "throw";
          context.arg = record.arg;
        }
      }
    };
  }

  // Call delegate.iterator[context.method](context.arg) and handle the
  // result, either by returning a { value, done } result from the
  // delegate iterator, or by modifying context.method and context.arg,
  // setting context.delegate to null, and returning the ContinueSentinel.
  function maybeInvokeDelegate(delegate, context) {
    var method = delegate.iterator[context.method];
    if (method === undefined) {
      // A .throw or .return when the delegate iterator has no .throw
      // method always terminates the yield* loop.
      context.delegate = null;

      if (context.method === "throw") {
        if (delegate.iterator.return) {
          // If the delegate iterator has a return method, give it a
          // chance to clean up.
          context.method = "return";
          context.arg = undefined;
          maybeInvokeDelegate(delegate, context);

          if (context.method === "throw") {
            // If maybeInvokeDelegate(context) changed context.method from
            // "return" to "throw", let that override the TypeError below.
            return ContinueSentinel;
          }
        }

        context.method = "throw";
        context.arg = new TypeError(
          "The iterator does not provide a 'throw' method");
      }

      return ContinueSentinel;
    }

    var record = tryCatch(method, delegate.iterator, context.arg);

    if (record.type === "throw") {
      context.method = "throw";
      context.arg = record.arg;
      context.delegate = null;
      return ContinueSentinel;
    }

    var info = record.arg;

    if (! info) {
      context.method = "throw";
      context.arg = new TypeError("iterator result is not an object");
      context.delegate = null;
      return ContinueSentinel;
    }

    if (info.done) {
      // Assign the result of the finished delegate to the temporary
      // variable specified by delegate.resultName (see delegateYield).
      context[delegate.resultName] = info.value;

      // Resume execution at the desired location (see delegateYield).
      context.next = delegate.nextLoc;

      // If context.method was "throw" but the delegate handled the
      // exception, let the outer generator proceed normally. If
      // context.method was "next", forget context.arg since it has been
      // "consumed" by the delegate iterator. If context.method was
      // "return", allow the original .return call to continue in the
      // outer generator.
      if (context.method !== "return") {
        context.method = "next";
        context.arg = undefined;
      }

    } else {
      // Re-yield the result returned by the delegate method.
      return info;
    }

    // The delegate iterator is finished, so forget it and continue with
    // the outer generator.
    context.delegate = null;
    return ContinueSentinel;
  }

  // Define Generator.prototype.{next,throw,return} in terms of the
  // unified ._invoke helper method.
  defineIteratorMethods(Gp);

  Gp[toStringTagSymbol] = "Generator";

  // A Generator should always return itself as the iterator object when the
  // @@iterator function is called on it. Some browsers' implementations of the
  // iterator prototype chain incorrectly implement this, causing the Generator
  // object to not be returned from this call. This ensures that doesn't happen.
  // See https://github.com/facebook/regenerator/issues/274 for more details.
  Gp[iteratorSymbol] = function() {
    return this;
  };

  Gp.toString = function() {
    return "[object Generator]";
  };

  function pushTryEntry(locs) {
    var entry = { tryLoc: locs[0] };

    if (1 in locs) {
      entry.catchLoc = locs[1];
    }

    if (2 in locs) {
      entry.finallyLoc = locs[2];
      entry.afterLoc = locs[3];
    }

    this.tryEntries.push(entry);
  }

  function resetTryEntry(entry) {
    var record = entry.completion || {};
    record.type = "normal";
    delete record.arg;
    entry.completion = record;
  }

  function Context(tryLocsList) {
    // The root entry object (effectively a try statement without a catch
    // or a finally block) gives us a place to store values thrown from
    // locations where there is no enclosing try statement.
    this.tryEntries = [{ tryLoc: "root" }];
    tryLocsList.forEach(pushTryEntry, this);
    this.reset(true);
  }

  runtime.keys = function(object) {
    var keys = [];
    for (var key in object) {
      keys.push(key);
    }
    keys.reverse();

    // Rather than returning an object with a next method, we keep
    // things simple and return the next function itself.
    return function next() {
      while (keys.length) {
        var key = keys.pop();
        if (key in object) {
          next.value = key;
          next.done = false;
          return next;
        }
      }

      // To avoid creating an additional object, we just hang the .value
      // and .done properties off the next function object itself. This
      // also ensures that the minifier will not anonymize the function.
      next.done = true;
      return next;
    };
  };

  function values(iterable) {
    if (iterable) {
      var iteratorMethod = iterable[iteratorSymbol];
      if (iteratorMethod) {
        return iteratorMethod.call(iterable);
      }

      if (typeof iterable.next === "function") {
        return iterable;
      }

      if (!isNaN(iterable.length)) {
        var i = -1, next = function next() {
          while (++i < iterable.length) {
            if (hasOwn.call(iterable, i)) {
              next.value = iterable[i];
              next.done = false;
              return next;
            }
          }

          next.value = undefined;
          next.done = true;

          return next;
        };

        return next.next = next;
      }
    }

    // Return an iterator with no values.
    return { next: doneResult };
  }
  runtime.values = values;

  function doneResult() {
    return { value: undefined, done: true };
  }

  Context.prototype = {
    constructor: Context,

    reset: function(skipTempReset) {
      this.prev = 0;
      this.next = 0;
      // Resetting context._sent for legacy support of Babel's
      // function.sent implementation.
      this.sent = this._sent = undefined;
      this.done = false;
      this.delegate = null;

      this.method = "next";
      this.arg = undefined;

      this.tryEntries.forEach(resetTryEntry);

      if (!skipTempReset) {
        for (var name in this) {
          // Not sure about the optimal order of these conditions:
          if (name.charAt(0) === "t" &&
              hasOwn.call(this, name) &&
              !isNaN(+name.slice(1))) {
            this[name] = undefined;
          }
        }
      }
    },

    stop: function() {
      this.done = true;

      var rootEntry = this.tryEntries[0];
      var rootRecord = rootEntry.completion;
      if (rootRecord.type === "throw") {
        throw rootRecord.arg;
      }

      return this.rval;
    },

    dispatchException: function(exception) {
      if (this.done) {
        throw exception;
      }

      var context = this;
      function handle(loc, caught) {
        record.type = "throw";
        record.arg = exception;
        context.next = loc;

        if (caught) {
          // If the dispatched exception was caught by a catch block,
          // then let that catch block handle the exception normally.
          context.method = "next";
          context.arg = undefined;
        }

        return !! caught;
      }

      for (var i = this.tryEntries.length - 1; i >= 0; --i) {
        var entry = this.tryEntries[i];
        var record = entry.completion;

        if (entry.tryLoc === "root") {
          // Exception thrown outside of any try block that could handle
          // it, so set the completion value of the entire function to
          // throw the exception.
          return handle("end");
        }

        if (entry.tryLoc <= this.prev) {
          var hasCatch = hasOwn.call(entry, "catchLoc");
          var hasFinally = hasOwn.call(entry, "finallyLoc");

          if (hasCatch && hasFinally) {
            if (this.prev < entry.catchLoc) {
              return handle(entry.catchLoc, true);
            } else if (this.prev < entry.finallyLoc) {
              return handle(entry.finallyLoc);
            }

          } else if (hasCatch) {
            if (this.prev < entry.catchLoc) {
              return handle(entry.catchLoc, true);
            }

          } else if (hasFinally) {
            if (this.prev < entry.finallyLoc) {
              return handle(entry.finallyLoc);
            }

          } else {
            throw new Error("try statement without catch or finally");
          }
        }
      }
    },

    abrupt: function(type, arg) {
      for (var i = this.tryEntries.length - 1; i >= 0; --i) {
        var entry = this.tryEntries[i];
        if (entry.tryLoc <= this.prev &&
            hasOwn.call(entry, "finallyLoc") &&
            this.prev < entry.finallyLoc) {
          var finallyEntry = entry;
          break;
        }
      }

      if (finallyEntry &&
          (type === "break" ||
           type === "continue") &&
          finallyEntry.tryLoc <= arg &&
          arg <= finallyEntry.finallyLoc) {
        // Ignore the finally entry if control is not jumping to a
        // location outside the try/catch block.
        finallyEntry = null;
      }

      var record = finallyEntry ? finallyEntry.completion : {};
      record.type = type;
      record.arg = arg;

      if (finallyEntry) {
        this.method = "next";
        this.next = finallyEntry.finallyLoc;
        return ContinueSentinel;
      }

      return this.complete(record);
    },

    complete: function(record, afterLoc) {
      if (record.type === "throw") {
        throw record.arg;
      }

      if (record.type === "break" ||
          record.type === "continue") {
        this.next = record.arg;
      } else if (record.type === "return") {
        this.rval = this.arg = record.arg;
        this.method = "return";
        this.next = "end";
      } else if (record.type === "normal" && afterLoc) {
        this.next = afterLoc;
      }

      return ContinueSentinel;
    },

    finish: function(finallyLoc) {
      for (var i = this.tryEntries.length - 1; i >= 0; --i) {
        var entry = this.tryEntries[i];
        if (entry.finallyLoc === finallyLoc) {
          this.complete(entry.completion, entry.afterLoc);
          resetTryEntry(entry);
          return ContinueSentinel;
        }
      }
    },

    "catch": function(tryLoc) {
      for (var i = this.tryEntries.length - 1; i >= 0; --i) {
        var entry = this.tryEntries[i];
        if (entry.tryLoc === tryLoc) {
          var record = entry.completion;
          if (record.type === "throw") {
            var thrown = record.arg;
            resetTryEntry(entry);
          }
          return thrown;
        }
      }

      // The context.catch method must only be called with a location
      // argument that corresponds to a known catch block.
      throw new Error("illegal catch attempt");
    },

    delegateYield: function(iterable, resultName, nextLoc) {
      this.delegate = {
        iterator: values(iterable),
        resultName: resultName,
        nextLoc: nextLoc
      };

      if (this.method === "next") {
        // Deliberately forget the last sent value so that we don't
        // accidentally pass it on to the delegate.
        this.arg = undefined;
      }

      return ContinueSentinel;
    }
  };
})(
  // In sloppy mode, unbound `this` refers to the global object, fallback to
  // Function constructor if we're in global strict mode. That is sadly a form
  // of indirect eval which violates Content Security Policy.
  (function() { return this })() || Function("return this")()
);
});

/**
 * Copyright (c) 2014-present, Facebook, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

// This method of obtaining a reference to the global object needs to be
// kept identical to the way it is obtained in runtime.js
var g = (function() { return this })() || Function("return this")();

// Use `getOwnPropertyNames` because not all browsers support calling
// `hasOwnProperty` on the global `self` object in a worker. See #183.
var hadRuntime = g.regeneratorRuntime &&
  Object.getOwnPropertyNames(g).indexOf("regeneratorRuntime") >= 0;

// Save the old regeneratorRuntime in case it needs to be restored later.
var oldRuntime = hadRuntime && g.regeneratorRuntime;

// Force reevalutation of runtime.js.
g.regeneratorRuntime = undefined;

var runtimeModule = runtime;

if (hadRuntime) {
  // Restore the original runtime.
  g.regeneratorRuntime = oldRuntime;
} else {
  // Remove the global property added by runtime.js.
  try {
    delete g.regeneratorRuntime;
  } catch(e) {
    g.regeneratorRuntime = undefined;
  }
}

var index$5 = runtimeModule;

/**
 * EventEmitter 类，用来在客户端间通过 featureControl channel 传递消息
 */
var ControlCenter = function () {
	function ControlCenter() {
		classCallCheck(this, ControlCenter);

		this.init();
	}

	createClass(ControlCenter, [{
		key: 'init',
		value: function init() {
			var _this = this;

			// 初始化的方法， 创建一个 callback 的 hash 对象来存 callback，根据 timestamp 作为 id 来索引，一开始使用自增的 callbackId 导致只能 copilot 向 pilot 发送请求，所以改了。
			this.events = {};
			this.callbackId = 0;
			this.callbackMap = Object.create(null);
			// 监听来自 featureControl 的事件，如果 data 里的 id 能索引到 callback 则是自己发送过去的 request 收到的 回调，如果没有则是别的地方发送过来的 request
			mesh.eventManager.addEventHandler('featureControl', function (data) {
				var id = data.id;
				var cb = _this.callbackMap[id];
				if (cb) {
					// receive call back
					var err = data.error;
					cb(data, err);
				} else {
					// not a callback, receive request
					_this.handler(data);
				}
			});
		}
	}, {
		key: 'handler',
		value: function () {
			var _ref = asyncToGenerator(index$5.mark(function _callee(data) {
				var event, handler, result;
				return index$5.wrap(function _callee$(_context) {
					while (1) {
						switch (_context.prev = _context.next) {
							case 0:
								// 处理别的设备发过来的 request，所以这个事件必须是自己有定义过回调的，一般是在 FeatureManager 那边注册回调，也就是调用 event.on
								// 这里 toDeviceId 之后需要修改，根据多机操作的需求
								event = data.type;
								handler = this.events[event];

								if (!handler) {
									_context.next = 13;
									break;
								}

								_context.prev = 3;
								_context.next = 6;
								return handler(data);

							case 6:
								result = _context.sent;

								mesh.eventManager.sendEventToServer('featureControl', {
									data: result,
									type: event,
									id: data.id,
									missionId: mesh.missionManager.missionID,
									userId: mesh.accountManager.userID,
									toDeviceId: data.fromDeviceId
								});
								_context.next = 13;
								break;

							case 10:
								_context.prev = 10;
								_context.t0 = _context['catch'](3);

								mesh.eventManager.sendEventToServer('featureControl', {
									error: _context.t0.message,
									data: {},
									type: event,
									id: data.id,
									missionId: mesh.missionManager.missionID,
									userId: mesh.accountManager.userID,
									toDeviceId: data.fromDeviceId
								});

							case 13:
							case 'end':
								return _context.stop();
						}
					}
				}, _callee, this, [[3, 10]]);
			}));

			function handler(_x) {
				return _ref.apply(this, arguments);
			}

			return handler;
		}()

		/**
   *  注册一个事件回调
   * @param {*} event 
   * @param {*} handler 
   */

	}, {
		key: 'on',
		value: function on(event, handler) {
			this.events[event] = handler;
		}

		/**
   *  发送一个事件请求
   * @param {*} type - type of event being sent over `featureControl` channel
   * @param {*} data - data being sent over `featureControl` channel
   * @param {*} deviceId - id of device the event is sent to
   */

	}, {
		key: 'send',
		value: function send(type, data, deviceId) {
			var _this2 = this;

			return new Promise(function (resolve, reject) {
				var id = Date.now();
				mesh.eventManager.sendEventToServer('featureControl', {
					type: type,
					id: id,
					data: data,
					missionId: mesh.missionManager.missionID,
					userId: mesh.accountManager.userID,
					toDeviceId: deviceId ? deviceId : mesh.missionManager.releatedDeviceID,
					fromDeviceId: mesh.missionManager.currentActiveDeviceID
				}, function (err) {
					if (err) {
						reject(err);
						return;
					}
					// 发送成功后将时间戳作为 key 把回调存起来, 在 18 行 init 方法内调用
					_this2.callbackMap[id] = function (data, err) {
						if (err) {
							reject(err);
						} else {
							resolve(data.data);
						}
					};
				});
			});
		}
	}]);
	return ControlCenter;
}();

var event$2 = new ControlCenter();

var en = {
  'cancel': 'Cancel',
  'confirm': 'Confirm',
  'notice': 'Notice',
  'ok': 'OK',
  'apply': 'Apply',
  'delete': 'Delete',
  'execute_feature_request_title': 'Execute',
  'execute_feature_request_message': 'Allow executing a feature on your drone?',
  'execute_feature_request_confirm': 'Execute',
  'execute_feature_rejected': 'Execution rejected',
  'notice_page_back': 'Sure to leave?',
  'notice_page_done': 'Sure to done this feature?',
  'notice_already_have_feature_executing': 'Already have feature executing, want to stop it?',
  'feature_take_off_in_the_air_error': 'The aircraft has taken off',
  // camera
  'lens_setting': 'Lens Settings',
  'iso': 'ISO',
  'shutter': 'Shutter',
  'ev': 'EV',
  'white_balance': 'White Balance',
  'filter': 'Color',
  'filter_none': 'None',
  'filter_art': 'Art',
  'filter_black_white': 'Black and White',
  'filter_bright': 'Bright',
  'filter_d_cinelike': 'D-Cinelike',
  'filter_protrait': 'Portrait',
  'filter_m31': 'M31',
  'filter_kdx': 'kDX',
  'filter_prismo': 'Prismo',
  'filter_jugo': 'Jugo',
  'filter_d_log': 'D-Log',
  'filter_true_color': 'True Color',
  'filter_inverse': 'Inverse',
  'filter_reminiscence': 'Reminiscence',
  'filter_solarize': 'Solarize',
  'filter_posterize': 'Posterize',
  'filter_whiteboard': 'Whiteboard',
  'filter_blackboard': 'Blackboard',
  'filter_aqua': 'Aqua',
  'filter_delta': 'Delta',
  'filter_dk79': 'DK79',
  'filter_vision4': 'Vision4',
  'filter_vision6': 'Vision6',
  'filter_true_color_ext': 'True Color Ext',
  'filter_film_a': 'Film A',
  'filter_film_b': 'Film B',
  'filter_film_c': 'Film C',
  'filter_film_d': 'Film D',
  'filter_film_e': 'Film E',
  'filter_film_f': 'Film F',
  'filter_film_g': 'Film G',
  'filter_film_h': 'Film H',
  'filter_film_i': 'Film I',
  'saturation': 'Saturation',
  'contrast': 'Contrast',
  'sharpness': 'Sharpness',
  'video_resolution_and_frameRate': 'Resolution and Framerate',
  'video_file_format': 'Video Format',
  'video_standard': 'Video Standard',
  'photo_format': 'Photo Format',
  'photo_aspect_ratio': 'Image Size',
  'photo_mode': 'Photo Mode',
  'photo_mode_single': 'Single Shot',
  'photo_mode_single_type_single': 'Single Mode',
  'photo_mode_single_type_hdr': 'HDR Mode',
  'photo_mode_burst': 'Multiple',
  'photo_mode_AEB': 'AEB',
  'photo_mode_interval': 'Interval'
};

var zh = {
  'cancel': '取消',
  'confirm': '确认',
  'notice': '注意',
  'ok': '好',
  'apply': '应用',
  'delete': '删除',
  'execute_feature_request_title': '执行操作',
  'execute_feature_request_message': '允许执行操作？',
  'execute_feature_request_confirm': '执行',
  'execute_feature_rejected': '执行被拒',
  'notice_page_back': '确定离开？',
  'notice_page_done': '确定完成？',
  'notice_already_have_feature_executing': '已有功能在执行中，需要将它停止吗？',
  'feature_take_off_in_the_air_error': '飞行器已起飞',
  // camera
  'lens_setting': '镜头设置',
  'iso': 'ISO',
  'shutter': '快门',
  'ev': '曝光值',
  'white_balance': '白平衡',
  'filter': '颜色',
  'filter_none': '无',
  'filter_art': '艺术',
  'filter_black_white': '黑白',
  'filter_bright': '亮度',
  'filter_d_cinelike': 'D-Cinelike',
  'filter_protrait': '肖像',
  'filter_m31': 'M31',
  'filter_kdx': 'kDX',
  'filter_prismo': 'Prismo',
  'filter_jugo': 'Jugo',
  'filter_d_log': 'D-Log',
  'filter_true_color': '写真',
  'filter_inverse': '反转',
  'filter_reminiscence': '怀旧',
  'filter_solarize': '曝光',
  'filter_posterize': 'Posterize',
  'filter_whiteboard': '白板',
  'filter_blackboard': '黑板',
  'filter_aqua': 'Aqua',
  'filter_delta': 'Delta',
  'filter_dk79': 'DK79',
  'filter_vision4': 'Vision4',
  'filter_vision6': 'Vision6',
  'filter_true_color_ext': '写真 Ext',
  'filter_film_a': '电影 A',
  'filter_film_b': '电影 B',
  'filter_film_c': '电影 C',
  'filter_film_d': '电影 D',
  'filter_film_e': '电影 E',
  'filter_film_f': '电影 F',
  'filter_film_g': '电影 G',
  'filter_film_h': '电影 H',
  'filter_film_i': '电影 I',
  'saturation': '饱和度',
  'contrast': '对比度',
  'sharpness': '锐度',
  'video_resolution_and_frameRate': '分辨率及帧率',
  'video_file_format': '视频格式',
  'video_standard': '视频标准',
  'photo_format': '照片格式',
  'photo_aspect_ratio': '图像比例',
  'photo_mode': '拍照模式',
  'photo_mode_single': '单张拍摄',
  'photo_mode_single_type_single': '单拍模式',
  'photo_mode_single_type_hdr': 'HDR 模式',
  'photo_mode_burst': '连拍',
  'photo_mode_AEB': 'AEB',
  'photo_mode_interval': '间隔'
};

var languages = {
	en: en,
	zh: zh
};

/**
 *  一个简单的 i18n 方法
 * @param {} lang 
 */
function intl(lang) {
	// const locale = mesh.lang
	var locale = mesh.language;
	locale = locale.slice(0, 2).toLowerCase();
	var all = {};
	Object.keys(languages).map(function (item) {
		all[item] = Object.assign({}, languages[item], lang[item]);
	});
	return all[locale];
}

var lang = intl({});

/**
 * Feature Manager
 * 
 * @export
 * @class FeatureManager
 */

var FeatureManager = function () {
	function FeatureManager() {
		classCallCheck(this, FeatureManager);

		this.store = null;

		this.events = {
			startExecuteFeature: 'startExecuteFeature',
			addClientEventListener: 'addClientEventListener'
		};
		// if the user has a device with it in the mission
		// i.e. `mesh.missionManager.deviceID` is not null
		// it should be listening to the related events
		if (mesh.aircraft && mesh.aircraft.deviceID) {
			this._registHandler();
			this.registHandler();
		}
		this._registMissionFeatureControls();
		this.registerGetFeatureDataEvent();
	}

	/**
  * register common event handler
  * 
  * @memberof FeatureManager
  */


	createClass(FeatureManager, [{
		key: '_registHandler',
		value: function _registHandler() {
			var _this = this;

			event$2.on('getFeatureStatus', function () {
				return _this.getFeatureStatus();
			});

			event$2.on('startExecuteFeature', function (data) {
				var _data$data = data.data,
				    featureData = _data$data.featureData,
				    uuid = _data$data.uuid;


				if (!mesh.aircraft || !mesh.aircraft.deviceID) {
					throw mesh.log("Error handling startExecuteFeature: no aircraft is connected");
				}

				return _this.execute(featureData, uuid, mesh.aircraft.deviceID);
			});

			event$2.on('getItemFromMeshStorage', function (data) {
				if (!data || !data.data) {
					throw mesh.log("Error in getItemFromMeshStorage: invalid data");
				}

				if (mesh.aircraft && mesh.aircraft.deviceID) {
					return _this.getItemFromMeshStorage(data.data.key, mesh.aircraft.deviceID);
				}
			});

			event$2.on('setItemInMeshStorage', function () {
				var _ref = asyncToGenerator(index$5.mark(function _callee(data) {
					var _data$data2, key, item;

					return index$5.wrap(function _callee$(_context) {
						while (1) {
							switch (_context.prev = _context.next) {
								case 0:
									if (!(!data || !data.data)) {
										_context.next = 2;
										break;
									}

									throw mesh.log("Error in setItemInMeshStorage: invalid data");

								case 2:
									if (!(mesh.aircraft && mesh.aircraft.deviceID)) {
										_context.next = 5;
										break;
									}

									_data$data2 = data.data, key = _data$data2.key, item = _data$data2.item;
									return _context.abrupt('return', _this.setItemInMeshStorage(key, item, mesh.aircraft.deviceID));

								case 5:
								case 'end':
									return _context.stop();
							}
						}
					}, _callee, _this);
				}));

				return function (_x) {
					return _ref.apply(this, arguments);
				};
			}());

			event$2.on('deleteItemFromMeshStorage', function () {
				var _ref2 = asyncToGenerator(index$5.mark(function _callee2(data) {
					var key;
					return index$5.wrap(function _callee2$(_context2) {
						while (1) {
							switch (_context2.prev = _context2.next) {
								case 0:
									if (!(!data || !data.data)) {
										_context2.next = 2;
										break;
									}

									throw mesh.log("Error in deleteItemFromMeshStorage: invalid data");

								case 2:
									if (!(mesh.aircraft && mesh.aircraft.deviceID)) {
										_context2.next = 5;
										break;
									}

									key = data.data.key;
									return _context2.abrupt('return', _this.deleteItemFromMeshStorage(key, mesh.aircraft.deviceID));

								case 5:
								case 'end':
									return _context2.stop();
							}
						}
					}, _callee2, _this);
				}));

				return function (_x2) {
					return _ref2.apply(this, arguments);
				};
			}());
		}

		/**
   * regist feature control event, for handler page
   * 
   * @memberof FeatureManager
   */

	}, {
		key: '_registMissionFeatureControls',
		value: function _registMissionFeatureControls() {
			var featureControls = [{
				event: 'featureWillStart',
				handler: 'onStart'
			}, {
				event: 'featureDidStart',
				handler: 'onStartSuccess'
			}, {
				event: 'featureStartFail',
				handler: 'onStartFail'
			}, {
				event: 'featureDidPause',
				handler: 'onPauseSuccess'
			}, {
				event: 'featurePauseFail',
				handler: 'onPauseFail'
			}, {
				event: 'featureDidResume',
				handler: 'onResumeSuccess'
			}, {
				event: 'featureResumeFail',
				handler: 'onResumeFail'
			}, {
				event: 'featureDidStop',
				handler: 'onStopSuccess'
			}, {
				event: 'featureStopFail',
				handler: 'onStopFail'
			}, {
				event: 'featureDidFinish',
				handler: 'onFinish'
			}, {
				event: 'featureExecuteFail',
				handler: 'onFail'
			}, {
				event: 'featureExecutionProgress',
				handler: 'onExecutionProgress'
			}, {
				event: 'activeDeviceChanged',
				handler: 'onActiveDeviceChanged'
			}];

			featureControls.map(function (control) {
				mesh.eventManager.addEventHandler(control.event, function (data) {
					var i = Route.stack.length - 1;
					while (i >= 0) {
						var lastStack = Route.stack[i];
						if (lastStack.isHandlerPage || lastStack.isMainPage) {
							if (isFunction(lastStack.instance.props[control.handler])) {
								lastStack.instance.props[control.handler](data || {});
							}
							return;
						} else {
							i--;
						}
					}
					mesh.log('can not find handler page');
				});
			});
		}

		/**
   * Set store for retrieving feature data
   * 
   * @param {any} store - store holding all feature data on current mobile device
   */

	}, {
		key: 'setStore',
		value: function setStore(store) {
			this.store = store;
		}

		/**
   * Register event for getting feature data
   * 
   */

	}, {
		key: 'registerGetFeatureDataEvent',
		value: function registerGetFeatureDataEvent() {
			var _this2 = this;

			event$2.on("featureData", function () {
				var _ref3 = asyncToGenerator(index$5.mark(function _callee3(data) {
					var _data$data3, mapKey, deviceId;

					return index$5.wrap(function _callee3$(_context3) {
						while (1) {
							switch (_context3.prev = _context3.next) {
								case 0:
									_data$data3 = data.data, mapKey = _data$data3.mapKey, deviceId = _data$data3.deviceId;

									if (!(_this2.store && _this2.store[mapKey] && _this2.store[mapKey][deviceId])) {
										_context3.next = 7;
										break;
									}

									_context3.next = 4;
									return _this2.getFeatureData(deviceId, mapKey);

								case 4:
									return _context3.abrupt('return', _context3.sent);

								case 7:
									throw mesh.log("Error when getting featureData, featureDataMap not available, or the given deviceId is invalid");

								case 8:
								case 'end':
									return _context3.stop();
							}
						}
					}, _callee3, _this2);
				}));

				return function (_x3) {
					return _ref3.apply(this, arguments);
				};
			}());
		}

		/**
   * Get feature data for a given device
   * 
   * @param {string} deviceId - id of device we are querying for
   * @param {string} mapKey - key of the feature data map to be retrieved from 
   */

	}, {
		key: 'getFeatureData',
		value: function () {
			var _ref4 = asyncToGenerator(index$5.mark(function _callee4(deviceId, mapKey) {
				var data;
				return index$5.wrap(function _callee4$(_context4) {
					while (1) {
						switch (_context4.prev = _context4.next) {
							case 0:
								if (!(this.store && this.store[mapKey] && this.store[mapKey][deviceId])) {
									_context4.next = 4;
									break;
								}

								return _context4.abrupt('return', this.store[mapKey][deviceId]);

							case 4:
								_context4.next = 6;
								return event$2.send("featureData", { mapKey: mapKey, deviceId: deviceId }, deviceId);

							case 6:
								data = _context4.sent;
								return _context4.abrupt('return', data);

							case 8:
							case 'end':
								return _context4.stop();
						}
					}
				}, _callee4, this);
			}));

			function getFeatureData(_x4, _x5) {
				return _ref4.apply(this, arguments);
			}

			return getFeatureData;
		}()

		/**
   * 调用 feature
   * 
   * @param {any} featureData - a map between device id and feature data
   * @param {any} uuid - the generated uuid to identify current handling
   * @param {any} toDeviceId - id of device executing features
   * 
   * @memberof FeatureManager
   */

	}, {
		key: 'execute',
		value: function execute(featureData, uuid, toDeviceId) {
			if (mesh.aircraft && toDeviceId === mesh.aircraft.deviceID) {
				mesh.featureExecutor.executeFeatures(featureData, uuid);
			} else {
				this._executeFeatureInCoPilot(featureData, uuid, toDeviceId);
			}
		}

		/**
   * execute feature in copilot
   * 
   * @param {any} featureObject - the feature data object
   * @param {any} uuid - the generated uuid to identify current handling
   * @param {any} toDeviceId - id of device this featureData is sent to
   * 
   * @returns 
   * @memberof FeatureManager
   */

	}, {
		key: '_executeFeatureInCoPilot',
		value: function _executeFeatureInCoPilot(featureData, uuid, toDeviceId) {
			var _this3 = this;

			return event$2.send('startExecuteFeature', {
				featureData: featureData,
				uuid: uuid
			}, toDeviceId).then(function () {}).catch(function (err) {
				_this3.handleError(err);
			});
		}

		/**
   * Delete an item with a given key in a device 
   * 
   * @param {any} key - key of the item to be deleted
   * @param {any} deviceId - id of device we are operating
   */

	}, {
		key: 'deleteItemFromMeshStorage',
		value: function () {
			var _ref5 = asyncToGenerator(index$5.mark(function _callee5(key, deviceId) {
				return index$5.wrap(function _callee5$(_context5) {
					while (1) {
						switch (_context5.prev = _context5.next) {
							case 0:
								if (!(!deviceId || mesh.aircraft && mesh.aircraft.deviceID && deviceId === mesh.aircraft.deviceID)) {
									_context5.next = 4;
									break;
								}

								mesh.storage.removeItem(key);
								_context5.next = 6;
								break;

							case 4:
								_context5.next = 6;
								return event$2.send('deleteItemFromMeshStorage', {
									key: key
								}, deviceId);

							case 6:
							case 'end':
								return _context5.stop();
						}
					}
				}, _callee5, this);
			}));

			function deleteItemFromMeshStorage(_x6, _x7) {
				return _ref5.apply(this, arguments);
			}

			return deleteItemFromMeshStorage;
		}()

		/**
   * Retrieve an item from mesh storage
   * 
   * @param {string} key - key of the item
   * @param {string} deviceId - id of device we are operating
   */

	}, {
		key: 'getItemFromMeshStorage',
		value: function () {
			var _ref6 = asyncToGenerator(index$5.mark(function _callee6(key, deviceId) {
				var paths, data;
				return index$5.wrap(function _callee6$(_context6) {
					while (1) {
						switch (_context6.prev = _context6.next) {
							case 0:
								if (!(!deviceId || mesh.aircraft && mesh.aircraft.deviceID && deviceId === mesh.aircraft.deviceID)) {
									_context6.next = 5;
									break;
								}

								paths = mesh.storage.getItem(key) || [];
								return _context6.abrupt('return', paths);

							case 5:
								_context6.next = 7;
								return event$2.send('getItemFromMeshStorage', {
									key: key
								}, deviceId);

							case 7:
								data = _context6.sent;
								return _context6.abrupt('return', data);

							case 9:
							case 'end':
								return _context6.stop();
						}
					}
				}, _callee6, this);
			}));

			function getItemFromMeshStorage(_x8, _x9) {
				return _ref6.apply(this, arguments);
			}

			return getItemFromMeshStorage;
		}()

		/**
   * Set an item under mesh storage
   * 
   * @param {string} key - key of the history data
   * @param {any} item - item to be set
   * @param {string} deviceId - id of device we are operating
   */

	}, {
		key: 'setItemInMeshStorage',
		value: function () {
			var _ref7 = asyncToGenerator(index$5.mark(function _callee7(key, item, deviceId) {
				return index$5.wrap(function _callee7$(_context7) {
					while (1) {
						switch (_context7.prev = _context7.next) {
							case 0:
								if (!(!deviceId || mesh.aircraft && mesh.aircraft.deviceID && deviceId === mesh.aircraft.deviceID)) {
									_context7.next = 4;
									break;
								}

								mesh.storage.setItem(key, item);
								_context7.next = 6;
								break;

							case 4:
								_context7.next = 6;
								return event$2.send('setItemInMeshStorage', {
									key: key,
									item: item
								}, deviceId);

							case 6:
							case 'end':
								return _context7.stop();
						}
					}
				}, _callee7, this);
			}));

			function setItemInMeshStorage(_x10, _x11, _x12) {
				return _ref7.apply(this, arguments);
			}

			return setItemInMeshStorage;
		}()

		/**
   * alert error
   * 
   * @param {any} err 
   * @memberof FeatureManager
   */

	}, {
		key: 'handleError',
		value: function handleError(err) {
			var buttons = [{ 'title': lang['ok'], 'style': 'blue' }];
			mesh.alert(lang['notice'], err.message || err, null, buttons, function () {});
		}
	}, {
		key: 'addClientEventListener',
		value: function addClientEventListener(eventName, handler) {
			if (mesh.aircraft && mesh.aircraft.deviceID) {
				mesh.eventManager.addEventHandler(eventName, function (data) {
					handler && handler(data);
					event$2.send(eventName, data);
				});
			} else {
				event$2.send('addClientEventListener', {
					event: eventName
				});
				event$2.on(eventName, function (data) {
					handler(data.data);
				});
			}
		}
	}]);
	return FeatureManager;
}();

/**
 * 由于 jsEngine 未暴露全局 global 或者 window 对象，编译时直接索引 this 会被
 * rollup 视为错误，所以只能通过这样来获取全局对象
 */
var globalObject = function () {
	return this;
}();

/**
 * 在全局对象中暴露一些接口，供客户端调用
 */
// polyfill 
globalObject.Promise = es6Promise.Promise;

globalObject.moduleId = 0;
globalObject.GestureEventHandler = GestureEventHandler;
globalObject.requireModule = requireModule;
globalObject.destroyComponent = destroyComponent;
globalObject.pageWillShow = pageWillShow;
registNativeModule();

if (!isBrowser) {
	globalObject.document = index$1;
}

mesh.log('Build Version: ' + process.env.BUILD_VERSION);

var index = {
	EMPTY_OBJ: EMPTY_OBJ,
	MODULE_MAP: MODULE_MAP,
	event: event$2,
	linkEvent: linkEvent,
	patchQueue: patchQueue,
	excutePatchQueue: excutePatchQueue,
	render: render,
	testRender: testRender,
	createVNode: createVNode
};

export { EMPTY_OBJ, MODULE_MAP, linkEvent, event$2 as event, patchQueue, excutePatchQueue, createVNode, testRender, render, FeatureManager, intl };export default index;
